/** ===========================================================================
	* ObjIn
	*	=====
	* Input/deserialization.
	* Class-specific readObject() and writeObject() serialization.
	* USAGE:
	* 	java JSX.ObjIn < test.xml
	**/

//package ObjOutDev;
package JSX;
//import JSX.*;	//use this, not "package", if truly a "wrapper"
import java.io.*;	//Stream

public class ObjIn extends ObjectInputStream {
	//static final boolean DEBUG = false;
	static final boolean DEBUG = true;

	//static final boolean TRACE = false;
	static final boolean TRACE = true;



	/**	==========================================================================
		* CONSTRUCTORS
		* ============
		**/
	/** -------------------------------------------------------------------------
		* ObjIn(in) - core constructor (others call it).
		* ------
		**/
	private XMLDeserialize d;
	public ObjIn(InputStreamReader in) throws IOException {
		super();	//sets flag, so readObject() defers to overrideReadObject()
		d = new XMLDeserialize(in);
		d.setArg(new Object[] {this});
			//d.readObjectArglist = new Object[] {this};
		if (TRACE) System.err.println("ObjIn(inReader)");
	}

	/** -------------------------------------------------------------------------
		* ObjIn(in)
		* ------
		* Convenience: wraps as a Reader.
		**/
	public ObjIn(InputStream in) throws IOException {	//concession to convenience
		this(new InputStreamReader(in));
		if (TRACE) System.err.println("ObjIn(in)");
	}

	/** -------------------------------------------------------------------------
		* ObjIn()
		* ------
		* Default to System.in (so can redirect files in etc).
		*	Incidentally: SecurityException is RunTime, so needn't explicitly throw.
		**/
	protected ObjIn() throws IOException, SecurityException {
		this(System.in);		//default
		if (TRACE) System.err.println("ObjIn()");
	}



	/**	==========================================================================
		* METHODS
		*	=======
		**/
	/** -------------------------------------------------------------------------
		* reset()
		* -------
		*	To reset alias table, mainly.
		* Not needed: resets automatically upon each new invocation.
		* Consider: option to change this when common objects are repeatedly
		* 	messaged.
		* Question: is this in OIS API?
		**/
	public void reset() {	
		d.reset();
	}

	/** -------------------------------------------------------------------------
		* readObjectOverride() -- different arg/return form output version
		* --------------------
		*	The heart.  ObjectInputStream.readObject() defers to this, if flag set.
		* Consider: only IOException should be thrown.
		**/
	protected Object readObjectOverride() throws IOException {
		if (TRACE) System.err.println("readObjectOverride()");
		Object o = null;
		try {
			o = d.deserialize();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return o;
	}

	/** -------------------------------------------------------------------------
		* defaultReadObject()
		* -----------------
		* Not implemented: presently, default stuff is written anyway.
		**/
	public void defaultReadObject() throws IOException {
		if (TRACE) System.err.println("defaultReadObject()");
	}
	public int readInt() throws IOException {
		return Integer.parseInt(d.getPrimString());
	}
	public float readFloat() throws IOException {
		return Float.parseFloat(d.getPrimString());
	}
		//we are stricter than Java's own parse of Boolean...
	public boolean readBoolean() throws IOException {
		String s = d.getPrimString();
		if (s.equals("false")) return false;
		else if (s.equals("true")) return true;
		else
			throw new IOException("Boolean must be 'true' or 'false'; not '"
				+s+"'");
	}
	public byte readByte() throws IOException {
		return Byte.parseByte(d.getPrimString());
	}
	public char readChar() throws IOException {
		//XML...
		return d.getPrimString().charAt(0);	//Question: is there a nicer parse?
	}
	public double readDouble() throws IOException {
		return Double.parseDouble(d.getPrimString());
	}
	//Float - see above
	//Int - see above
	public short readShort() throws IOException {
		return Short.parseShort(d.getPrimString());
	}
		//Also thanks to Andy Thomas - odd that both were left out!
	public long readLong() throws IOException {
		return Long.parseLong(d.getPrimString());
	}



	/**	==========================================================================
		* TEST CODE
		* =========
		**/
	public static void main(String[] args) throws Exception {
		ObjIn in = new ObjIn();
		ObjOut out = new ObjOut();
		Object o;
		while ( (o = in.readObject()) != null) {	//loop for streamed XML
			if (DEBUG) {	//sep debug info from deserialization and serialization
				System.err.println();
				System.err.println("===============================================================================");
				System.err.println();
			}
			out.writeObject(o);
		}
	}
}
