/** TITLE: XMLDeserialize.java (JSX0.8).
	*        -------------------
	*
	* Creates the object described by an XML document.

	* Note that a basic limitation of Java is that an object cannot be created
	* unless the JVM has access to the class file defining it.  That is, Java
	* cannot *define* classes at runtime; it can only instantiate pre-defined
	* classes.
	* However, a class loader could be written to "read" in class files which
	* do not actually exist; thus giving Java the ability to define new
	* classes dynamically.  See also JavaClass (link on JSX homepage).

	* Please email me: for help, to suggest features, or to report bugs.
		bren@csse.monash.edu.au  <Brendan Macmillan>

	* ISSUES:
	*		What about: subclassing wrappers or containers?
	*		What if root object is a wrapper?  Or a field is a wrapper?
	
	**/


package JSX;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;	//for Hashtable

public class XMLDeserialize {

/** --------------------------------------------------------------------------
	* debug constants
	* ---------------
	**/
	private static final boolean DEBUG = false;
	//private static final boolean DEBUG = true;

	private static final boolean TESTCIRC = false;
	//private static final boolean TESTCIRC = true;

	private static final boolean ALIASDEBUG = false;
	//private static final boolean ALIASDEBUG = true;

	private static final boolean INTERNAL_DEBUG = false;
	//private static final boolean INTERNAL_DEBUG = true;

/** --------------------------------------------------------------------------
	* Deserialization fields
	* ----------------------
	* NB: there are others scattered through the code, proximate to their use.
	**/
	private ParserXML p;

/** --------------------------------------------------------------------------
	* alias fields
	* ------------
	* Consider: extract as an object.  It has state, init, and API.
	* Consider: change name from 'aliasHash' to 'aliasTable'.
	**/
	private int aliasSerialNumber;	//NB: *not* static (cs serialization)
	private Hashtable aliasHash = new Hashtable();
		//private static final int ALIAS_INIT = -1;//so pre-increment in serializeObject
		//get INIT directly from XMLSerialize, so guaranteed to be synchronized.


	
	/**--------------------------------------------------------------------------
		* main
		* ----
		* Version information only; for test code, see JSXTest.java
		**/
	public static void main(String args[]) {
		System.err.println(XMLSerialize.VERSION);	
	}



/**==========================================================================
	* CONSTRUCTORS
	* ============
	**/
	/**--------------------------------------------------------------------------
		* XMLDeserialize(in) - core constructor
		* --------------
		* Core constructor (others call it)
		* init alias *only* here, or confuses re-enterant code (internal ser)
		**/
	public XMLDeserialize(Reader in) {	
		reset();	//whenever create, always reset alias table.
		p = new ParserXML(in);
	}
	//--------------------------------------------------------------------------
	/** XMLDeserialize()
		* --------------
		* Default to System.in
		**/
	public XMLDeserialize() {			//default to System.in
		this(new InputStreamReader(System.in));
	}



/**==========================================================================
	* METHODS
	* =======
	**/
	int invocationCount = 0;
	//--------------------------------------------------------------------------
	/** deserialize()
		* -----------
		* Entry point.
		* NB: invocationCount distinguisges between first time callers, and
		* calls from internal serialization.  This allows both old and new
		* API's to work.
		**/
	
	public Object deserialize() throws Exception {
if(DEBUG) System.out.println("deserialize()");
		if (invocationCount==0)
			reset();
		invocationCount++;
		ParserXML.Tag tag = p.readTag();
		if (tag==null) return null;		//no more objects left
		if (!tag.start)
			throw new ParserXML.ExceptionXML("Expected start tag, got \"</" +tag.name+ ">\"");
if(DEBUG) System.out.println("1");
		Object o = createObject(tag.name, null);	//no parent
if(DEBUG) System.out.println("2");
		if (TESTCIRC)
			System.err.println("Count was: " + (aliasSerialNumber-XMLSerialize.ALIAS_INIT));
			if (ALIASDEBUG) System.err.println(aliasHash);
		invocationCount--;
		return o;	//separate statement into 2 lines so we can print the count
	}

	//NB: same as XMLSerialize
	public void reset() {
		if (invocationCount!=0)
			throw new Error("Deserialiize: Attempt to reset alias table mid-way.");
		initAlias();
	}



	/**--------------------------------------------------------------------------
		* alias nascent-object
		* --------------------
		* NB: init, fields and API: extract into a class.
		**/
	private void initAlias() {
		aliasSerialNumber = XMLSerialize.ALIAS_INIT;
		aliasHash.clear();
	}
	/**--------------------------------------------------------------------------
		* put Alias
		* --------
		*	inline = !serialized;
		* if (inline)
		*		//add to hash
		*	else
		*		//write as alias
		*	return (inline);		//to tell caller to serialize it inline!
		*		//else, caller leaves the ref writing up to this routine
		**/
	//private static boolean putAlias(Object o, String fieldName, boolean isStringAttr, String indent) {
	/**--------------------------------------------------------------------------
		* getAlias
		* --------
		* Much of this is duplicated code from addAttr().  This should be
		*	rationalized and refactored at some point.
		**/
	private Object getAlias(String flag, Object parent) throws ParserXML.ExceptionXML, Exception {
		ParserXML.Attr attr;
		String fieldName = null;
		String aliasName = null;
		if (flag.equals(XMLSerialize.ALIAS_TAG_TOKEN)) {	//parse obj-name and name
			while ( !(attr=p.readAttr()).isEnd) {
				if (attr.name.equals(XMLSerialize.NAME_TOKEN)) {
					if (fieldName!=null) 
						throw new ParserXML.ExceptionXML("just one name field");
					fieldName = attr.value;
				}
				else if (attr.name.equals(XMLSerialize.ALIAS_ATTR_TOKEN)) {
					if (aliasName!=null)
						throw new ParserXML.ExceptionXML("just one alias name");
					aliasName = attr.value;		//NB: won't flag duplicates
				} else
						throw new ParserXML.ExceptionXML("unknown attribute: "+attr.name);
			}
 			Object me = aliasHash.get(aliasName);
			if (parent!=null) {	//ie, if not an orphan object (as in Vector, Array)
				Field f = getAllField(parent.getClass(), fieldName);
				f.set(parent, me);
			}
				//System.err.println("alias " +aliasName+ ":");
				//System.err.println("me: " +me);
 			return me;
		}
		else
			return null;
	}

	private void putAlias(Object o) {
		if (o==null) return;	//ignore
		++aliasSerialNumber;	//pre-inc (use this syntax to highlight it)
		aliasHash.put(aliasSerialNumber+"", o);	//add (NB: inverse of serialize)
		if (TESTCIRC) System.err.println("inline " +aliasSerialNumber+ ":");
/*
		Object alias;	//will be String, but save the conversion
		if ( (alias = aliasHash.get(o))==null) {
			++aliasSerialNumber;	//pre-inc (use this syntax to highlight it)
			aliasHash.put(o, aliasSerialNumber+"");	//add (convert to String)
				//System.err.println(indent +"inline " +aliasSerialNumber+ ":");
			//Need to:
			//	indicate "inline" somehow.
			//	write as inline (same as now).
			return true;	//yes, should serialize the object inline
		}
		else
		{
			if (isStringAttr) {
				printAttr(fieldName + STRING_ALIAS_TOKEN, alias+"");
			}
			else {
					//System.err.println(indent+ "alias " +alias+ ":");
				printOpenTag(ALIAS_TAG_TOKEN, indent);
				if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);	//nice to reuse
				printAttr(ALIAS_ATTR_TOKEN, alias+"");
				out.println("/>");		//empty tag - under control of caller
			}
			return false;	//no, don't serialize inline; we (or you?) write the alias
		}
*/
	}



	static final Class[] OIS_ARGS = {ObjectInputStream.class};
	public Object[] readObjectArglist;
	/**--------------------------------------------------------------------------
		* setArg()
		* -------
		* Consider: putting this in constructor; or caller in same package or same
		* class.
		**/
	public void setArg(Object[] argList) {
		readObjectArglist = argList;
	}
	Vector primStore; //Consider: best place to define?
	int primStoreIndex;
	/**--------------------------------------------------------------------------
		* getPrimString()
		* ---------------
		* Called by readInt(), readFloat(), etc to get the String that was
		* stored by addAttr
		**/
	public String getPrimString() {
		return (String) primStore.get(primStoreIndex++);
	}
	/**--------------------------------------------------------------------------
		* createObject()
		* ------------
		* Returns the deserialized object (caller needs to cast at runtime)
		* The object created is "me".
		* USAGE: the open tag is assumed to have been just read
		* LOGIC: A new object is instantiated: "me".  Attr and elements arex
		* added to me.  This new object is then added to its parent.  NB: that
		* the name for me (given by parents) is only discoverable by reading the
		* attr; which is done in "addAttr" (second arg is boolean for isArray)
		* NB: a null object may still have a name.
		* NOTE: tag problem for lengths - not actually a field.
		**/
	private Object createObject(String elementName, Object parent)
			throws Exception {
		Object me;
		Class clazz;
if(DEBUG) System.out.println("createObject("+elementName+", "+parent+")");
	//mapping from my conventions to actual classes...
		if (elementName.equals(XMLSerialize.NULL_TOKEN)) {
			me = null;		//[null] NB: uses code at end!  to read end.
										//I think, to cater for
										//(1). reading rest of open tag AH! need name attr...
										//(2). including empty tag/close tag
							//Nice to reuse tested code, but this is awkward & confusing...
			clazz = null;
			//ParserXML.Attr name = addAttr(me, false);	//gets NAME_TOKEN
			//addTags(me);	//I *think* needed to get to end
			//return me;
				//TRY THIS: I may have missed some intricacy; but easier to debug
				//in short: too much if/else and complex logic.
				//an interesting approach: just save, then try to simplify it *as is*
		}
		else if (elementName.startsWith(XMLSerialize.ARRAY_TOKEN)) {
			me = createArray(elementName, parent);	//need length here...
			//QUESTION: shouldn't this also be put into the alias table?
			//NB: if so, also need to do so in serialize
			//Consider: arrays of arrays is tricky to specify (inc is OK)
			return me;		//better to do all else internally (inc name & parent)
			//NB: it *must* read right up to this close of this tag
			//NB: correct nesting of array types is not checked.
		}
                else if (elementName.startsWith(XMLSerialize.PROXY_TOKEN)) {
		    me = createProxy(elementName, parent);
                    return me;
                }
                else if (elementName.startsWith(XMLSerialize.CLASS_TOKEN)) {
		    me = createClass(elementName, parent);
                    return me;
                }
		else if ( (me=getAlias(elementName, parent))!=null )
			return me;
		else {	//or ignore rest
			//me = Class.forName(elementName).newInstance();//only new_Instance in file
			clazz = Class.forName(elementName);	//sep into 2 lines for debugging
/*
			if (me.getClass() == String.class) {	//TEST
				me = deserializeWrapper(elementName);
				putAlias(me);
				return me;
			}
*/
				//	me = clazz.newInstance();//exception from newInstance()
			//Constructor con = clazz.getConstructor(new Class[] {});	//ie no-arg con
			//con.setAccessible(true);
			//me 	= con.newInstance(new Object[] {});	//no-arg again
			if (clazz==Vector.class || clazz==Hashtable.class)	//special cases
				me = clazz.newInstance();
			else
				me = MagicClass.newInstance(clazz);	//use anon
			putAlias(me);	//inline=true
		}
		//NB: an array may also have a name!
		//[null] NB: me may be null!  (set above)
		Vector localStackPrimStore = primStore; //save locally, so recurse
		int localStackPrimStoreIndex = primStoreIndex;	//also save it.
			//Consider: use an explicit stack?
			//(assuming that recursion is the best solution)
			//NB: we store locally, before addAttr set primStore to another value
			//NB: primStore needs to be an obj var, so readInt() etc have access
		ParserXML.Attr name = addAttr(me, false);	//adds in all attr and gets NAME_TOKEN
		//Test: what's in this Vector?
			if (INTERNAL_DEBUG) {
				if (primStore!=null) {
					for (Enumeration keys=primStore.elements(); keys.hasMoreElements();) {
						Object a = keys.nextElement();
						System.err.println("In primStore: " +a+ " -- " + a.getClass() + " -- " +a.hashCode());
					}
				}
			}

			//inside: if me is an array, don't add no fields!

		if (DEBUG) System.err.println("Made class \"" +elementName+ "\"");
			//needs to be after "me" is created...
			//but array needs the length before can be created - circular dependancy
			//best to fix while rationalizing the fields.
		//or, allow arrays to handle their attributes separately - this is correct.

		if (parent!=null) {		//root doesn't have a parent
			Field f = getAllField(parent.getClass(), name.value);
			//System.err.println("About to set field \"" +name.value+ "\"");
			f.set(parent, me);
		}		//awkward meanings: name==null; and name.emptyTag
		if (DEBUG)
			System.err.println("\n\nGot an: \n"+  name);
	
//	 if (!name.emptyTag) {	//dup because of logic change in maintenance
		if (me!=null && me.getClass() == Vector.class) {
	 		if (!name.emptyTag) {	//dup because of logic change in maintenance
				addVectorElements((Vector) me);	//(could poly it.)
			}
		}
		else if (me!=null && me.getClass() == Hashtable.class) {
	 		if (!name.emptyTag) {	//dup because of logic change in maintenance
				addHashtableElements((Hashtable) me);
				//name.length;		//HACK - message passing
			}
		}
		else {
			//WE ONLY CALL THIS *AFTER* we have read all the information in.
			//... how do we know how much info to read in?
			//a better design would be to load all the objects into a Vector,
			//along with distguishing info: name for field, no name (or number)
			//for internal objects (actually, in different vectors).
			//The problem with this is that we can't just read the top level
			//of the tags in - we must parse them as we find them, and they
			//may recurse considerably.  Taking this approach leads us to make
			//a whole in-memory DOM, in effect.  Actually, perhaps we do do
			//that, but on the stack...?
			//NB: we put the internal objects first, because know when they end
			//ASSUME: (adapted from code/comments in XMLSerialize)
			//(0). We ignore Hashtable and Vector (code flow does this already)
			//(1). defaultReadObject() does nothing (or we read header twiceget double header)
			//(2). readObject() does not do primitives - only objects. Attr!
			//(3). we need to reset alias hash - do it manually for now.
			//still experimental !!  Determining effect of position...
			if (me!=null) {
				Method m;
				Vector classes = new Vector();	//NB: Mirror order - reverse.
				for (Class c=clazz; c!=Object.class; c=c.getSuperclass()) {//no f==nullA
//					classes.add(c);
//				}	//Consider: nicer to create Vector in reverse order, first?
//					//Even nicer: use a stack to reverse things... some deepness here?
//				for (int i=classes.size()-1; i>=0; i--) {	//step backwards
//					Class c = (Class) classes.get(i);
					if ((m=XMLSerialize.getDeclaredMethod(c, "readObject", OIS_ARGS,
							Modifier.PRIVATE, Modifier.STATIC)) != null) {	//NB: clazz->c
						if (INTERNAL_DEBUG) {
							System.err.print("\t--JSX deserialize--");
						}
						try {
							m.invoke(me, readObjectArglist);	//might as well do them all!
							if (INTERNAL_DEBUG) {
								System.err.println("\tinvoking: " +m);
								System.err.println("\ton: " +me);
								System.err.println("\tof: " +clazz);
								System.err.println("\tHERE");
							}
						} catch (Exception e) {
							System.err.println(e);
						}
					}
				} //for superclass loop
			}
	 		if (!name.emptyTag) {	//dup because of logic change in maintenance
				addTags(me);
			}
		}
	 //}        
		primStoreIndex = localStackPrimStoreIndex;	//restore
		primStore = localStackPrimStore;//restore primitives for previous level
		return me;
	}


	/**--------------------------------------------------------------------------t
		* getAllField()
		* -----------
		* Named in accord with the JSX serialization method "getAllFields()".
		* returns private and superclass fields.
		*  NOTE: could subclass Field instead.  More natural, as we are really
		*  extending its functionality.  However, it is also convenient to keep
		*  it in the one class.
		**/
	private	Field getAllField(Class c, String fieldName) throws NoSuchFieldException, SecurityException {
		Field f = null;
		Class initc = c;	//store it for error
		do {
			try {
				f = c.getDeclaredField(fieldName);
				f.setAccessible(true);		//so can see private fields (Java 1.2)
			} catch (NoSuchFieldException e) {	//do nothing; it's OK
			} catch (Exception e) {
				System.err.println(e +"No such field '"+ fieldName +"' in "+ c +" (superclas of " +initc);
			}
		} while ( f==null && (c = c.getSuperclass()) != Object.class);
		if (f==null)
			throw new NoSuchFieldException("JSX can't find the field "
			+((fieldName==null)?"<null>":"'"+fieldName+"'")
			+ ".  Not in " +initc+ ", or private, or in superclasses\n"
			+ "Could it be that the class lacks a readObject() method?");
		return f;
	}


/**==========================================================================**/
/**======================= ARRAY SECTION ====================================**/
/**==========================================================================**/

	/**--------------------------------------------------------------------------
		* createArray
		* -----------
		* Create one Array object.
		* Note that the whole name must be parsed to know the component type,
		* which may be: (ARRAY_TOKEN)*Type

		* Essentially, maps (ARRAY_TOKEN)*Type -> Type([])*
		* It counts the ARRAY_TOKENS, extracts the type, then:
		*  Array.new_Instance(Class.getClass(baseType), new int[count])

		* NB: we avoid the complex disambiguation constraint on tok, by name choice
		**/
	private Object createArray(String tagName, Object parent) throws Exception {
		//System.err.println("ARRAY DESERIALIZATION NOT YET IMPLEMENTED (properly)");
		//As I say, gentle reader, there is lots of awful hacking here - but think!
		//- that is just how Michael Angelo brought forth David.  He hacked at
		//several blocks of stone, gaining understanding with each, before
		//completing the final masterpiece.  I am at the still at the first block,
		//and I've learnt a lot about the shape of the problem from it!
		// - Brendan "hacker or sculpter? you decide!" Macmillan

		//NOTE!! need this here, so we know length
		ParserXML.Attr attr = addAttr(null, true); //to get name and length

			//NB: called twice!  (for arrays of primitives):
			//first: to get length (and name, too)
			//second: to get elements.
			//	thus, we need to store the name til after the element
			//	NB: if name came after the elements, no problem
			//	we could make it order independent.

			//attr==null; attr.emptyTag; attr.length
			//NB!! addAttr should *only* read the name and length: there is no point
			//in reading the element values themselves, because the array itself has
			//not yet been created.

			//This implies that the length *must* be given before any element values.
			//It doesn't logically preclude the name to be given at any point, but
			//the present organisation of the program expects it also to be given
			//before any of the element values.

			//THUS, we need
			// (1). addAttr to return after seeing "length"; and
			// (2). to call addAttr a second time, treating attr as indices.

			//MORE problems!  addAttr needs to behave differently depending or
			//whether it is of primitive or of reference base type.  If primitive,
			//it returns immediately on reading the length;  if reference, it
			//swollows up the ">", at the end of the tag.
			//SOLN: because an array of reference type can never contain any
			//attr but name and length, we can additionally read these in, in a
			//hack.  All this must be rewritten cleanly, now that all (I think)
			//consequences of the data-binding are apparent - or perhaps a rewrite
			//of the data-binding might be more appropriate  ;-)
		int dim = 0;
		int index=0;
		String tok = XMLSerialize.ARRAY_TOKEN;
		int skip = tok.length();
		//(1). take off the leading "ArrayOf-"'s (by incrementing an index for eff)
		while (tagName.startsWith(tok, index)) { //could be > eff if assume tok qual
			dim++;														//and just check first char...
			index+=skip;
		}
		if (DEBUG) System.err.println("dim= " +dim);
		String type = tagName.substring(index);	//get the type
		if (DEBUG) System.err.println("type= " +type);
		Object me = null;
		try {
			int[] dimsArray = new int[dim];
			dimsArray[0] = attr.length;		//NOTE use of length parsed from attributes
			Class comp;
			//shockingly inefficient!  This is probably why they used the 'L'
			if (type.equals("int")) comp = int.class;
			else if (type.equals("double")) comp = double.class;
			else if (type.equals("boolean")) comp = boolean.class;
			else if (type.equals("char")) comp = char.class;
			else if (type.equals("byte")) comp = byte.class;
			else if (type.equals("short")) comp = short.class;
			else if (type.equals("long")) comp = long.class;
			else if (type.equals("float")) comp = float.class;
			else comp = Class.forName(type);
			me = Array.newInstance(comp, dimsArray);
			putAlias(me);
		} catch (Exception e) {
			System.err.println("Illegal type as Array base component, it seems; "+e);
		}
		if (DEBUG) System.err.println("type= " +me.getClass().getName());
		if (DEBUG) System.err.println("len= " +Array.getLength(me));

		ParserXML.Tag tag;
		if (dim==1 && me.getClass().getComponentType().isPrimitive()) {
			//read all the following attrs into the array
			String rememberFieldName = attr.value;	//because called twice. YUK
			attr = addAttr(me, true); //NB: length is not needed again by this route
			if (attr.value==null) attr.value = rememberFieldName;
				//the "if" allows us to accept "obj-name" before or after elements.
			if (DEBUG) System.err.println("after setting primitive array, saw: " + attr);
		}	//attr (for primitive base type)
		else {
			//do the contained objects (as does Vector)
			//NOTE  the recursion runs through create object, which tests isArray,
				//so as to automatically bottom out.
				//AND WHAT A HACK IT IS!
			ParserXML.Attr end = p.readAttr(); //slurp up the left-over '>' after length read.
			end.length = attr.length;
			if (DEBUG) System.err.println("after starting reference array, saw: " + end);
				//NB: this implies that the name (if present) *must* precede length
			for (int i=0; i<end.length; i++) {
				//System.err.println("i = "+i);
				Object obj;
		 		tag=p.readTag();
				if (tag==null) throw new Exception("Premature EOF");
				if (!tag.start)
					throw new Exception("Expected "+end.length+" elements in array; read only"+i);
				if (tag.name.equals("java.lang.String")	//bad: doubles up
						|| tag.name.equals("java.lang.Integer")
						|| tag.name.equals("java.lang.Double")
						|| tag.name.equals("java.lang.Boolean")
						|| tag.name.equals("java.lang.Character")
						|| tag.name.equals("java.lang.Byte")
						|| tag.name.equals("java.lang.Short")
						|| tag.name.equals("java.lang.Long")
						|| tag.name.equals("java.lang.Float")
					)
					obj = deserializeWrapper(tag);
				else
					obj = createObject(tag.name, null);	//null=orphan object
				if (DEBUG) System.err.println("Adding {"+obj+"} to Array");
				Array.set(me, i, obj);
			}	//for loop
		}	//tags (for refernce base type - array or class)
		if (!attr.emptyTag) {		//nice!
			tag=p.readTag();
			if (tag==null) throw new Exception("Premature EOF");
			if (tag.start)
				throw new Exception("End of array");
			if (!tag.name.equals(tagName))
				throw new ParserXML.ExceptionXML(
					"</"+tagName+">","</"+tag.name+">");
		}
			//if get here, all is OK!

		if (parent!=null) {		//root doesn't have a parent CODE DUP
			if (attr.value==null)
				throw new Exception("No " +XMLSerialize.NAME_TOKEN+ " attribute found for "+parent.getClass());
			Field f = getAllField(parent.getClass(), attr.value);
			f.set(parent, me);
		}		//misuse as message gives awkward meanings:

		return me;
	}


    private Object createProxy(String tagname, Object parent) throws Exception
    {
        p.readAttr();
        ParserXML.Tag tag=p.readTag();
        InvocationHandler handler=(InvocationHandler)createObject(tag.name, parent);
        tag=p.readTag();
        Class[] interfaces = (Class[])createObject(tag.name, parent);
        p.readTag();
        return Proxy.newProxyInstance(this.getClass().getClassLoader(), interfaces, handler);
    }

    private Object createClass(String tagname, Object parent) throws Exception
    {
        ParserXML.Attr attr=p.readAttr();
        p.readAttr();
        return Class.forName(attr.value);
    }

	/**--------------------------------------------------------------------------
		* getArrayType: "ArrayOf-int" to "LI"
		* ------------
		* convert Array name string from:
		*		[*		//any number of '[' (including 0)
		*		(B|C|D|F|I|J|S|Z) |		//one of these represeting a primitive type
		*   (L[^;]*;)		//or the name of the class, in between 'L' and ';'
		**/
/*
	private static String getArrayType(Class clazz) {
		String a = "";	//will hold result
		String name = clazz.getName();
		if (DEBUG) System.err.println("Array type = \""+name+"\"");
		int i;
		for (i=0; name.charAt(i)=='['; i++)
			a += ARRAY_TOKEN;
		switch (name.charAt(i++)) {
			case 'B': a += "byte";
				break;
			case 'C': a += "char";
				break;
			case 'D': a += "double";
				break;
			case 'F': a += "float";
				break;
			case 'I': a += "int";
				break;
			case 'J': a += "long";
				break;
			case 'S': a += "short";
				break;
			case 'Z': a += "boolean";
				break;
			case 'L': a += name.substring(i, name.lastIndexOf(';'));
						//could also be name.length()-1
			break;
		}
		// NB: in a multi-dimensional array, Class.getComponentType() returns
		// the next one further in (which may also be an array).
		return a;
	}
*/




	/**--------------------------------------------------------------------------t
		* addAttr (map to primitives and String)
		* -------
		* Only: int and String
		*	Should be: byte,char,short,int,long,float,double and String.
		*
		* BIZARRE: Could do something odd for arrays: take the object only as
		* indicating the exact type of the array, and in here, when read the
		* length, actually create it.
		*	And yet it makes sense in a way.
		*
		* or, return a proper "attrMsg", which includes the length as info.
		* This is wha should be returned, anyway - not this hacked use of Attr.
		**/
	private ParserXML.Attr addAttr(Object o, boolean isArray) throws Exception {
		int length = -1;	//to flag if not set
		ParserXML.Attr attr;
		ParserXML.Attr name = null;
		if (!isArray)	{ //array never has internal serialized values
			primStore = new Vector();		//create anew for each level (so can recurse)
			primStoreIndex = 0;	//refresh at same time.
		}
		while ( !(attr=p.readAttr()).isEnd ) {
			if (DEBUG) System.err.println(attr);
			if (attr.name.equals(XMLSerialize.NAME_TOKEN)) { //if name of this class
				if (name!=null)
					throw new ParserXML.ExceptionXML("just one name field");
				name = attr;
				continue;	//next attr
			}
				//ignore all fields of array, but for the name
			if (isArray) {	//length or primitive attributes
				if (attr.name.equals(XMLSerialize.LENGTH_TOKEN)) {	//if length
					if (INTERNAL_DEBUG) {
						System.err.println("ARRAY LENGTH: " +attr);
					}
					length = Integer.parseInt(attr.value);
					break;		//break off part way through (will this work?)
				}
				if (INTERNAL_DEBUG) {
					System.err.println("ARRAY PRIMITIVE ATTRIBUTES: " +attr);
				}
				//else, continues on, reading primitive values, and sticking them in.
				//assumed: .startsWith(XMLSerialize.ARRAY_PRIMITIVE_INDEX_TOKEN)
				String indexStr = attr.name.substring(XMLSerialize.ARRAY_PRIMITIVE_INDEX_TOKEN.length());
				int index = Integer.parseInt(indexStr);	//we know it is an int
				if (DEBUG) System.err.println(attr);
				Object value = getValue(o.getClass().getComponentType(), attr);		//hmmmph!  Seems quite simple
				Array.set(o, index, value);
			}
			else {	//if not array...
				if (attr.name.endsWith(XMLSerialize.ALIAS_STRING_TOKEN)) {//string alias
					if (INTERNAL_DEBUG) {
						System.err.println("STRING ALIAS: " +attr);
					}
					//(1). extract the fieldName
					String fieldName = attr.name.substring(0, attr.name.indexOf(XMLSerialize.ALIAS_STRING_TOKEN));
					//(2). get the value & use it to lookup the object
					Object strObj = aliasHash.get(attr.value);	//get the object (String)
					//(4). assign that object to the field
					Field f = getAllField(o.getClass(), fieldName);
					Class fc = f.getType();			//type specified by java, not by xml
					if (DEBUG) System.err.println("Made Field \""+fc.getName()+"\"");
					f.set(o, strObj);	//note: referring to already XML decoded object
				}
				else {	//if not string alias...
					if (attr.name.startsWith(XMLSerialize.INTERNAL_PRIMITIVE_TOKEN)) {
						if (INTERNAL_DEBUG) {
							System.err.println("INTERNAL: " +attr);
						}
						String primString = attr.name.substring(XMLSerialize.INTERNAL_PRIMITIVE_TOKEN.length());	//omit prefix (typically '_' or "_." or "someclass.")
						if (DEBUG) System.err.println(attr);
							//non-XML: we rely on the ordering of attributes
							//need (perhaps) to detect non-ordering, and at that time
							//start allowing for it....
							//or, we could detect the size, and add in nulls to match it.
							//And signal an error later.
							//NB: I think that Strings will always be done as tags, even
							//though attributes might be neater.
						primStore.add(attr.value);	//keep it in the universal form of "string"
							//Question: how are internal Strings stored?
							//It doesn't matter, except that they will be read with
							//readObject, and so should not be on the primStore...
					}
					else {	//if an ordinary field (not an array at all)
						if (INTERNAL_DEBUG) {
							System.err.println("ordinary: "+attr);
						}
						Field f = getAllField(o.getClass(), attr.name);	//subclass Field?
						Class fc = f.getType();			//type specified by java, not by xml
						if (DEBUG) System.err.println("Made Field \""+fc.getName()+"\"");
						f.set(o, getValue(fc, attr));
					}
				}
			}
		}
			//combine info from the name and final attr returned.  Awkward.
		attr.length = length;
		if (name==null) {
				//attr.value = null;	//some paths assume the name is here... ugly
			return attr;	//simple.  No name to return
		}
			//else, combine info from both (name and last attr)
		name.isEnd = attr.isEnd;
		name.emptyTag = attr.emptyTag;
		name.length = attr.length;
		return name;
	}


	/**--------------------------------------------------------------------------
		* getValue
		* ------------------
		* In: the class of the primitive; the attr (name-value pair)
		* Out: * object representing the primitive
		* Class may come from the "TYPE" field, or from array component type.
		* Used by:
		*		deserializeWrapper
		*		addAttr - note: an intermediate wrapper object is created redundantly
		*
		* It may be more efficient to not reuse this code.
		* when used for an attribute, it creates an Object needlessly;
		* however, such an object is required for Hashtable
		**/
	private Object getValue(Class fc, ParserXML.Attr attr) throws Exception {
		if (DEBUG) System.err.println("getting value of " +fc);
		if (fc.isPrimitive()) {     //PRIMITIVE
			if (fc==int.class) return new Integer(attr.value);	//most likely first
			else if (fc==double.class) return new Double(attr.value); //second pr
			else if (fc==boolean.class) {
				if (attr.value.equals("false"))
					return new Boolean (false);
				else if (attr.value.equals("true"))
					return new Boolean (true);
				else
					throw new Exception("Boolean must be 'true' or 'false'; not '"
						+attr.value+"'");
			}
			else if (fc==byte.class) return new Byte(attr.value);
			else if (fc==char.class) {
				if (attr.value.length()!=1) throw new Exception("Character data must be exactly one character long; instead we got: '" +attr.value+ "'");
				return new Character(attr.value.charAt(0));
			}
			else if (fc==short.class) return new Short(attr.value);
			//else if (fc==int.class) return new Integer(attr.value));
			else if (fc==long.class) return new Long(attr.value);
			else if (fc==float.class) return new Float(attr.value);
			else throw new ParserXML.ExceptionXML(
				"Unimplemented primitive: \"" +fc.getName()+ "\"");
		}
		else {                      //CLASS
			if (fc==String.class) {
				String s = decodeXML(attr.value);
				putAlias(s);	//String is an object - not called for attr string
				return s;
			}
			else throw new ParserXML.ExceptionXML(
				"'"+attr.name+"' is not primitive or String");
		}
	}



	/**--------------------------------------------------------------------------
		* addTags
		* -------
		* Handles just tags (no attrs)
		* CONSIDER: is there an elegant way to reuse this code in Hashtable?
		* This is where the end tag is read
		* Perhaps pass in the open tag explicity, rather than (inefficiently)
		* recalling it from the object itself.  Why?  Because this doesn't work
		* for null objects.
		**/
	private void addTags(Object me) throws Exception {
		ParserXML.Tag tag;
		while ( ((tag=p.readTag())!=null) && (tag.start))	
			createObject(tag.name, me);	//add this new tag to me
		if (tag==null)
			throw new ParserXML.ExceptionXML("Premature EOF");
	
		if (me==null) {		//dangling else!
			if (!tag.name.equals(XMLSerialize.NULL_TOKEN))	//special case of null
				throw new ParserXML.ExceptionXML(
					"</"+XMLSerialize.NULL_TOKEN+">","</"+tag.name+">");
		}
		else
			if (!tag.name.equals(me.getClass().getName()))
				throw new ParserXML.ExceptionXML(
					"</"+me.getClass().getName()+">","</"+tag.name+">");
	}


	/**--------------------------------------------------------------------------
		* addVectorElements
		* -----------------
		* Note: this should now be possible automatically, since we can read
		* private fields (and any internal serialization)
		* Note: Vector object already created when we get here.
		**/
	private void addVectorElements(Vector v) throws Exception {
		ParserXML.Tag tag;
		while (true) {
			Object obj;
			tag = p.readTag();
			if (tag==null) throw new Exception("Premature EOF");
			if (!tag.start) return;	//done!  caller will check if a correct endtag
			if (tag.name.equals("java.lang.String")	//bad: doubles up
					|| tag.name.equals("java.lang.Integer")
					|| tag.name.equals("java.lang.Double")
					|| tag.name.equals("java.lang.Boolean")
					|| tag.name.equals("java.lang.Character")
					|| tag.name.equals("java.lang.Byte")
					|| tag.name.equals("java.lang.Short")
					|| tag.name.equals("java.lang.Long")
					|| tag.name.equals("java.lang.Float")
				)
				obj = deserializeWrapper(tag);
			else
				obj = createObject(tag.name, null);	//null=orphan object
			if (DEBUG) System.err.println("Adding {"+obj+"} to Vector");
			v.add(obj);
		}
		//if (DEBUG) System.err.println("Vector: \n" +v);
	}

	/**--------------------------------------------------------------------------
		* addHashtableElements
		* --------------------
		* An container object.
		* Handles just tags (no attrs)
		**/
	private void addHashtableElements(Hashtable h) throws Exception {
		ParserXML.Tag tag;
		while (true) {
			Object[] objs = {null, null};
			for (int i=0; i<objs.length; i++) {		//{key, value}
				tag = p.readTag();
				if (tag==null) throw new Exception("Premature EOF");
				if (!tag.start)
					if (i==0) return; //end of hashtable
					else throw new Exception("incomplete Hashtable entry");
				if (tag.name.equals("java.lang.String")	//bad: doubles up
						|| tag.name.equals("java.lang.Integer")
						|| tag.name.equals("java.lang.Double")
						|| tag.name.equals("java.lang.Boolean")
						|| tag.name.equals("java.lang.Character")
						|| tag.name.equals("java.lang.Byte")
						|| tag.name.equals("java.lang.Short")
						|| tag.name.equals("java.lang.Long")
						|| tag.name.equals("java.lang.Float")
					)
					objs[i] = deserializeWrapper(tag);
				else
					objs[i] = createObject(tag.name, null);	//null=orphan object
			}
			if (DEBUG) System.err.println("Putting {"+objs[0]+", "+objs[1]+"}");
			h.put(objs[0], objs[1]);
		}
		//if (DEBUG) System.err.println("Hashtable: \n"+h);
	}


	/**--------------------------------------------------------------------------
		* deserializeWrapper
		* ------------------
		* Handles the attrs.
		* Need to refactor the test for wrapper type
		**/
	private Object deserializeWrapper(ParserXML.Tag startTag) throws Exception {
		ParserXML.Attr attr, end;
		ParserXML.Tag tag;
		attr = p.readAttr();
		if (!attr.name.equals("valueOf")) 
			throw new ParserXML.ExceptionXML("valueOf expected");
		if (!(end = p.readAttr()).isEnd)
			throw new ParserXML.ExceptionXML("\">\" or \"/>\" expected");
		if (!end.emptyTag) {
			tag = p.readTag();
			if (!tag.name.equals(startTag.name) || tag.start)
				throw new ParserXML.ExceptionXML("</"+tag.name+"> expected");
		}
		//IDEA: could make up the expected tag, and use Tag.equals() to compare
		if (DEBUG) System.err.println("String read in: " + attr.value);
		if (startTag.name.equals("java.lang.String")) {	
			String s = decodeXML((String)attr.value);	//not attr.name! Don't forget to decode!
			putAlias(s);	//String is an object
			return s;
		}
		Class c = Class.forName(startTag.name);
		try {
			Field f = getAllField(c, "TYPE");
			Object o = getValue((Class)f.get(c), attr);
			putAlias(o);	//NOTE: *not* called if a String, so not inc twice
			return o;
		} catch (NoSuchFieldException e) {
			System.err.println("'" +c+ "' is not a wrapper class: "+e);
		}
		throw new Exception("Shouldn't get here");
	}


	/**--------------------------------------------------------------------------
		* decodeXML
		* ---------
		* &amp; etc.
		* To do: this also needs to be applied to char (not just Strings)
		**/
	private String decodeXML(String in) {
		String a = "";
		int i;
		try {
			for (i=0; i<in.length(); i++) {
				char c = in.charAt(i);
				if (c=='&') {				//if the escape character
					i++;
					if (in.startsWith("amp;", i))       {a+='&';  i+=3;} //NB: loop's x++
					else if (in.startsWith("lt;", i))   {a+='<';  i+=2;}
					else if (in.startsWith("gt;", i))   {a+='>';  i+=2;}
					else if (in.startsWith("apos;", i)) {a+='\''; i+=4;}
					else if (in.startsWith("quot;", i)) {a+='"';  i+=4;}
				}
				else
					a += c;
			}
		} catch (IndexOutOfBoundsException e) {
			System.err.println( "Seems to be an incomplete escaped entity: "
				+in.substring(in.lastIndexOf('&')) );
		}
		return a;
	}


	/**--------------------------------------------------------------------------
		* Test class
		* ----------
		**/
	public static class Test {
		public static void main(String args[]) throws Exception {

			XMLDeserialize d = new XMLDeserialize(new InputStreamReader(System.in));
			Object o;
			while ( (o = d.deserialize()) != null) {	//can loop for streamed XML
			/*
				System.err.println("Printing created object:");
				System.err.println();
				System.err.println(o);
				System.err.println();
				System.err.println("(Re)serializing created object:");
				System.err.println();
			*/
				XMLSerialize.serialize(o);
			}
		}
	}
}
