package momoko;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.beans.*;
import momoko.straw.*;
import momoko.relationship.*;
import momoko.drivers.*;
import momoko.versioncontrol.*;
import momoko.protoclass.*;
import momoko.rpc.*;
import momoko.rmi.*;
import momoko.types.*;
import momoko.stream.*;

/**
 *  Holds all information about the known universe
 */
public class Database
{
  public boolean debug=false;
  boolean init=false;
  static public boolean client=false;
/**
 *  primary database instance
 */
//  public static Database main = new Database();
  public static Database main;
/**
 *  available primative types
 */
  public static Types types = new Types();
/**
 *  database of external device drivers
 */
  public DriverManager driver;
  public RemoteObjectServer remoteint;
  public LocalBroker broker;
  private Map remoteobjs = new Hashtable();
/**
 *  root of virtual file system
 */
  public IMObject root;
/**
 *  root of class directory
 */
  public IMObject classroot;
/**
 *  database of relationships between objects
 */
  public Map relationships;
/**
 *  shortcut to the containment relationship
 */
  public Relationship contains;
/**
 *  shortcut to the parent-child relationship
 */
  public Relationship parent;
/**
 *  ClassLoader to reload classes after compilation
 */
  static public AdaptiveClassLoader loader;

  static public Vector manifest;

  static private String hostname;
  static private int port;

/**
 *  Creates a new database which must be initialized before use
 */
public Database()
{
this(false);
}

public Database(boolean c)
{
  client=c;
try{
  File f=null;
//  f=new File("lib");
  Vector v=new Vector();
  if(f!=null && f.exists())
  {
    String[] libs=f.list();
    for(int x=0; x<libs.length; x++)
      v.addElement(new File("lib/"+libs[x]));
  }
  loader = new AdaptiveClassLoader(v);
  driver = new DriverManager("config/drivers.config");
  remoteint=new RMIServer();
  broker=new GenericStrawBroker();
  if(root==null)
  {
    if(debug) System.out.println("Regenerating core database.");
    root=new MObject("root");
  }
  if(relationships==null)
  {
    if(debug) System.out.println("Regenerating relationship database.");
    relationships=new Hashtable();
  }
  contains=(Relationship)relationships.get("contains");
  if(contains==null)
  {
    contains=new SubjectExclusiveRelationship("contains", true);
    contains.setfunctions(new DirectoryAddPredicate(), new DirectoryRemovePredicate());
    relationships.put("contains",contains);
  }
  parent=(Relationship)relationships.get("begets");
  if(parent==null)
  {
    parent=new SubjectExclusiveRelationship("parent", false);
    relationships.put("begets",parent);
  }
} catch(Exception e) {if(debug) System.out.println("Error in Database constructor:"); e.printStackTrace();}
}

/**
 *  regenerates user database and class directory
 */
public void initialize()
{
  if(init) return;
  classroot=(IMObject)regenerate("momoko.protoclass.GenericClassRepository",root,"class");
  IMObject world=(IMObject)regenerate("momoko.MObject",root,"world");
  regenerate("momoko.user.Admin",world,"God");
  init=true;
}

/**
 *  shut down database
 */
public void shutdown()
{
  unmountall();
  remoteint.shutdown();
}

public String getHostname()
{
  return hostname;
}

public void setHostname(String newHostname)
{
  hostname=newHostname;
}

public int getPort()
{
  return port;
}

public void setPort(int p)
{
  port=p;
}

/**
 *  find an object and create it if it doesn't exist
 */
public IMObject regenerate(String parent, IMObject location, String name)
{
  IMObject o;

//  if(debug) System.out.println("Attempting to regenerate "+parent+" - "+name);
  try { return resolve(location, name); }
  catch(BadPathException e)
  {
    if(debug) System.out.println("Regenerating "+name+"("+parent+")");
    return addobj(parent,location,name);
  }
}

/**
 *  display directory tree
 */
public void tree()
{
  if(debug) System.out.println("root");
  treeprint(root, " ");
}

public void tree(StreamHandler handler)
{
  handler.format("root");
  treeprint(handler, root, " ");
}

private void treeprint(IMObject obj, String padding)
{
  Map con;
  Iterator iterator;
  String s;

  if(debug) System.out.println(obj);
  con=obj.contents();
  iterator=con.keySet().iterator();
  while(iterator.hasNext())
  {
    s=(String)iterator.next();
    if(debug) System.out.println(padding+s);
    treeprint((IMObject)con.get(s), padding+" ");
  }
}

private void treeprint(StreamHandler handler, IMObject obj, String padding)
{
  Map con;
  Iterator iterator;
  String s;

  con=obj.contents();
  iterator=con.keySet().iterator();
  while(iterator.hasNext())
  {
    s=(String)iterator.next();
    handler.format(padding+s);
    Object o=con.get(s);
    treeprint(handler, (IMObject)o, padding+" ");
  }
}

public IMObject addobj(String parent, String location, String name) throws IllegalArgumentException
{
  IMObject obj, pobj=null;
  Enumeration iterator;
  Class parentclass;
  Constructor parentconstructor;
  Class argt[]=new Class[1];
  Object argo[] = {name};
//if(debug) System.out.println("addobj("+parent+", "+location+", "+name+")");
  if(parent==null) throw new IllegalArgumentException();
  if(location==null) throw new IllegalArgumentException();
  if(name==null) throw new IllegalArgumentException();
  try
  {
    pobj=((VersionController)resolve(classroot,parent)).getCurrent();
  }
  catch(BadPathException e)
  { if(debug) System.out.println("Couldn't find parent "+parent); }

  try
  {
//if(debug) System.out.println("CREATING object with name: "+name);
    obj = (IMObject)root.create((IProtoClass)pobj, location, name);
//if(debug) System.out.println("Created object: "+obj+" "+obj.getName());
//if(debug) System.out.println("PARENT "+pobj.getName()+" "+obj.getName());
    main.parent.relate(pobj,obj);
//    this.move(obj,location);
    return obj;
  }
  catch(Exception e)
  {
    if(debug) System.out.println("Database.addobj failed: "+e.getClass().getName());
    if(debug) System.out.println("Failed on p:"+parent+" - n:"+name+" - l:"+location);
    e.printStackTrace();
    return null;
  }
}

/**
 *  create a new object and store it in the directory tree
 */
public IMObject addobj(String parent, IMObject location, String name) throws IllegalArgumentException
{
  IMObject obj, pobj=null;
  Enumeration iterator;
  Class parentclass;
  Constructor parentconstructor;
  Class argt[]={String.class};
  Object argo[] = {name};

  if(parent==null) throw new IllegalArgumentException();
  if(location==null) throw new IllegalArgumentException();
  if(name==null) throw new IllegalArgumentException();
  try
  {
    if(debug) System.out.println("loading class "+parent);
    try { parentclass=Class.forName(parent); }
    catch(ClassNotFoundException a)
    {
     try { parentclass=Class.forName("momoko."+parent); }
     catch(ClassNotFoundException e)
     { 
       try
       {
        parentclass=loader.loadClass(parent);
        pobj=((VersionController)resolve(classroot,parent)).getCurrent();
       }
       catch(Exception g)
       {         
        if(debug) System.out.println("Error getting class or parent."); 
        g.printStackTrace(); return null;
       }
     }
    }
if(debug)System.out.println("getting constructor from "+parentclass.getName());
    parentconstructor=parentclass.getConstructor(argt);
    obj = (IMObject)parentconstructor.newInstance(argo);
    if(pobj!=null)
      main.parent.relate(pobj,obj);
    this.move(obj,location);
    return obj;
  }
  catch(java.lang.reflect.InvocationTargetException e)
  {
    e.getTargetException().printStackTrace();
    return null;
  }
  catch(Exception e)
  {
    if(debug) System.out.println("Database.addobj failed: "+e.toString());
    if(debug) System.out.println("Failed on "+parent+" - "+name+" - "+location);
    e.printStackTrace();
    return null;
  }
}

public IMObject addobj(IProtoClass pobj, IMObject location, String name) throws IllegalArgumentException
{
  IMObject obj;

  if(pobj==null) throw new IllegalArgumentException();
  if(location==null) throw new IllegalArgumentException();
  if(name==null) throw new IllegalArgumentException();
  String parent=pobj.name();
  try
  {
    obj=pobj.newInstance(name);
    if(obj==null)
    {
      if(debug) System.out.println("Could not make new obj for class "+pobj.getName());
      throw new IllegalArgumentException();
    }
//if(debug) System.out.println("addobj 3 trying to relate "+pobj+" "+obj);
    main.parent.relate(pobj,obj);
    this.move(obj,location);
    return (IMObject)obj;
  }
  catch(Exception e)
  {
    if(debug) System.out.println("Database.addobj failed: "+e.toString());
    if(debug) System.out.println("Failed on "+parent+" - "+name+" - "+location);
    e.printStackTrace();
    return null;
  }
}

/**
 *  destroy an object
 */
public void delobj(String path)
{
  try
  {
    IMObject victim=resolve(root, path);
    IMObject container=victim.location();
    contains.unrelate(container, victim);
    parent.unrelate(victim.getparent(), victim);
    Map contents=container.contents();
    contents.remove(victim);
    container.setcontents(contents);
  }
  catch(Exception e)
  {
    if(debug) System.out.println("Exception in delobj(): "+e);
  }
}

/**
 *  change an object's position in the directory tree and it's containment relationship
 */
public void move(IMObject obj, IMObject dest) throws ClassException
{
  Map lc, dc;
  String name=obj.name();
  IMObject location=obj.location();

  if(dest==null) return;
  if(location==dest) return;

  if(location!=null)
    if(!location.release(obj))
    {
      if(debug) System.out.println("Object not released!");
      return;
    }

  if(dest.accept(obj))
  {
//    if(debug) System.out.println("Calling exitfunc: "+location+" "+obj);
    if(location!=null) location.exitfunc(obj);
    contains.relateforced(dest, obj);
    if(location!=null) location.remove(obj.getName());
    dest.enterfunc(obj);
  }
  else
    if(debug) System.out.println("Move rejected.");
}

public IMObject resolve(IMObject start, String path) throws BadPathException
{
  IMObject o;
  int i;
  String prefix, suffix, temp;
  Map contents;
  Enumeration iterator;
  boolean done=false;

  if(path.length()==0)
    return start;
//  if(debug) System.out.println("will resolve: "+start);
//  if(debug) System.out.println("resolving: "+path+" - "+start.getClass().getName());
  if(path.charAt(0)=='/')
    return this.resolve(root,path.substring(1));
  else
    return start.resolve(path);
}

/**
 *  find an object based on a path
 */
public IMObject oldresolve(IMObject start, String path) throws BadPathException
{
  IMObject o;
  int i;
  String prefix, suffix, temp;
  Map contents;
  Iterator iterator;
  boolean done=false;

  if(path.length()==0)
    return start;
//  if(debug) System.out.println("will resolve: "+start);
//  if(debug) System.out.println("resolving: "+path+" - "+start.getClass().getName());
  if(path.charAt(0)=='/')
    return this.resolve(root,path.substring(1));
  i=path.indexOf('/');
  if(i==-1)
  {
    prefix=path;
    suffix="";
  }
  else
  {
    prefix = path.substring(0, i);
    suffix = path.substring(i+1);
  }
  contents=start.contents();
  iterator = contents.keySet().iterator();
  while(iterator.hasNext()&&!done)
  {
    temp=(String)iterator.next();
//    if(debug) System.out.println("resolve searching through "+temp+" - "+contents.get(temp).getClass().getName());
    if(prefix.equals(temp))
      done=true;
  }
  if(!done)
    throw new BadPathException(path);
  return this.resolve((IMObject)contents.get(prefix),suffix);
}
  
/**
 *  open a gate to an object on a remote server
 */
public IMObject mount(String path, String name, String remote)
{
  // Is this still useful? I don't know.
  return null;
}

/**
 *  close all gates
 */
public void unmountall()
{
  // Is this still useful? I don't know.
}

}
