package momoko.compiler;

import momoko.*;
import java.util.*;
import java.io.*;
import momoko.util.*;
import momoko.protoclass.*;

/**
 * 
 */
public class JavaSourceWriter extends OOSourceWriter
{
  PrintWriter out;

public JavaSourceWriter(PrintWriter pw)
{
  super();
  out=pw;
  file_extension=".java";
}

public void write(Object o)
{
System.out.println("writing object");
  if(o instanceof ProtoClassSourceObject)
    write((ProtoClassSourceObject)o);
  else
    out.print(o);
}

public void write(Property o)
{
  out.print(print(o));
}

public void write(IProtoClass pc) throws IOException
{
  write(new PearlSourceObject(pc));
}

public void write(ProtoClassSourceObject pso)
{
System.out.println("writing class");
  if(pso instanceof PearlSourceObject)
    write((PearlSourceObject)pso);
  else
    write(new PearlSourceObject(pso));
}

public void write(InterfaceSourceObject iso)
{
  write("interface");
}

public void write(PearlSourceObject pso)
{
System.out.println("writing pearl");
  Enumeration iterator;
  write(pso.getPackage());
  iterator=pso.getLibraries().elements();
  while(iterator.hasMoreElements()) write((Library)iterator.nextElement());
  write(new ClassHeader(pso.getName(), pso.getParent(), pso.getInterfaces()));
  iterator=pso.getVariables().elements();
  while(iterator.hasMoreElements()) write((Variable)iterator.nextElement());
  iterator=pso.getConstructors().elements();
  while(iterator.hasMoreElements()) write((Constructor)iterator.nextElement());
  iterator=pso.getMethods().elements();
  while(iterator.hasMoreElements()) write((Method)iterator.nextElement());
  write(new ClassFooter());
}

public void write(PackageDeclaration pkg)
{
System.out.println("set_scope: "+out);
      out.println("package "+pkg.pkg+";");
}

public void write(Library stmt)
{
  out.println("import "+stmt.pkg+"."+stmt.clss+";");
}

public void write(ClassHeader ch)
{
      out.print("public class "+ch.name+" extends "+ch.parent);
        Enumeration it=ch.interfaces.elements();
        if(it.hasMoreElements())
          out.print(" implements "+it.nextElement());
        while(it.hasMoreElements())
          out.print(", "+it.nextElement());
      out.println("{");
}

public void write(Variable iv)
{
//  out.println(Database.main.types.classname(iv.type)+" "+iv.name+";");
  out.println("public "+iv.type+" "+iv.name+";");
}

public void write(GetterMethod gm)
{
  Vector source=new Vector();
  source.addElement("return "+gm.name+";");
  write(new Method("get_"+gm.name, new Function(gm.type, new Vector(), source)));
}

public void write(SetterMethod gm)
{
  Vector source=new Vector();
  source.addElement(gm.name+"=new"+gm.name+";");
  Vector args=new Vector();
  args.addElement(new Pair(gm.type, "new"+gm.name));
  write(new Method("set_"+gm.name, new Function("void", args, source)));
}

public void write(Constructor c)
{
  write(new ConstructorHeader(c.name, c.args));
  write(new MethodBody(c.body));
  write(new ConstructorFooter());
}

public void write(ConstructorHeader ch)
{
        out.print("public "+ch.name+"(");
        Enumeration it=ch.args.elements();
        if(it.hasMoreElements())
        {
          p=(Pair)it.nextElement();
          out.print(Database.main.types.classname((String)p.first)+" "+p.second);
        }
        while(it.hasMoreElements())
        {
          p=(Pair)it.nextElement();
          out.print(", "+Database.main.types.classname((String)p.first)+" "+p.second);
        }
        out.println(")");
        out.println("{");
}

public void write(VarInit vi)
{
  write(new AssignmentStatement(vi.name,vi.value));
}

public void write(ConstructorFooter cf)
{
  write(new MethodFooter());
  out.println("\n");
}

public void write(Method m)
{
  if(m instanceof GetterMethod) { write((GetterMethod)m); return; }
  if(m instanceof SetterMethod) { write((SetterMethod)m); return; }
  write(m.header);
  write(m.body);
  write(m.footer);
}

public void write(MethodHeader mh)
{
//        out.print("public "+Database.main.types.classname(mh.type)+" "+mh.name+"(");
System.out.println("type "+mh.type);
        out.print("public "+mh.type+" "+mh.name+"(");
        Enumeration it=mh.args.elements();
        if(it.hasMoreElements())
        {
          p=(Pair)it.nextElement();
//          out.print(Database.main.types.classname((String)p.first)+" "+p.second);
          out.print(p.first+" "+p.second);
        }
        while(it.hasMoreElements())
        {
          p=(Pair)it.nextElement();
//          out.print(", "+Database.main.types.classname((String)p.first)+" "+p.second);
          out.print(", "+p.first+" "+p.second);
        }
        out.println(")");
        out.println("{");
}

public void write(MethodBody mb)
{
  Enumeration iterator=mb.elements();
  while(iterator.hasMoreElements())
    out.println(iterator.nextElement());
}

public void write(ReturnStatement rs)
{
  out.println("return "+rs.value+";");
}

public void write(AssignmentStatement as)
{
        out.println(as.to+"="+as.from+";");
}

public void write(MethodFooter mf)
{
        out.println("}\n");
}

public void write(ClassFooter cf)
{
System.out.println("class footer");
  out.println("}");
}

}
