package momoko.compiler;

import momoko.*;
import java.util.*;
import momoko.types.*;
import java.io.*;
import java.lang.String;

/**
 * 
 */
abstract public class OOCompiler extends Compiler
{
  public String file_extension="";
  protected IProtoClass pc;
  public PrintWriter out;
  protected Pair p;
  protected String s;
  protected Property prop;
  protected Enumeration iterator;

public File createfile(IProtoClass class_to_compile) throws IOException
{
      pc=class_to_compile;
      System.out.println("Creating file\n");
      if(pc==null) System.out.println("pc is null!");
      out=new PrintWriter(new FileWriter(pc.name()+file_extension));
System.out.println(out);

      IProtoClass parent=(IProtoClass)Database.main.parent.subject(pc);
      if(parent==null) throw new IOException("No parent");

      Vector libs=new Vector();
      libs.addElement(new Pair("java.util", "*"));
      libs.addElement(new Pair("mud.shared.types","MudArray"));
//      libs.addElement(new Pair("mud.shared","MudObject"));

      set_scope("mud.user");
      import_libraries(libs);
      class_header(pc.name(), parent.name());
      instance_variables(new Hashtable(pc.getproperties()));
      constructor(pc.name(), pc.getproperties());
      access_methods(pc.getproperties());
      create_method("create", (Property)pc.getproperties().get("create"));
      other_methods(pc.getproperties());
      class_footer();
      out.close();
      return new File(pc.name()+file_extension);
}

abstract public void set_scope(String scope);

public void import_libraries(Vector libs)
{
  Enumeration iterator=libs.elements();
  while(iterator.hasMoreElements())
  {
    Pair p=(Pair)iterator.nextElement();
    import_library((String)p.first, (String)p.second);
  }
}

abstract public void import_library(String path, String item);

abstract public void class_header(String classname, String parent);

public void instance_variables(Hashtable h)
{
      iterator=h.keys();
      while(iterator.hasMoreElements())
      {
        s=(String)iterator.nextElement();
        prop=(Property)h.get(s);
        if(prop.type.equals("Function")) continue;
        instance_variable(prop.type, s);
      }
}

abstract public void instance_variable(String type, String name);

public void constructor(String classname, Hashtable properties)
{
  constructor_header(classname);
  constructor_body(properties);
  constructor_footer();
}

abstract public void constructor_header(String classname);

public void constructor_body(Hashtable properties)
{
      iterator=properties.keys();
      while(iterator.hasMoreElements())
      {
        s=(String)iterator.nextElement();
        prop=(Property)properties.get(s);
        if(prop.type.equals("Function")) continue;
        initialize_instance_variable(s, prop);
      }
}

abstract public void initialize_instance_variable(String name, Property value);

public void constructor_footer()
{
      method_footer();
}

public void access_methods(Hashtable properties)
{
      iterator=properties.keys();
      while(iterator.hasMoreElements())
      {
        s=(String)iterator.nextElement();
        prop=(Property)properties.get(s);
        if(prop.type.equals("Function")) continue;
        get_method(prop.type, s);
        set_method(prop.type, s);
      }
}

public void get_method(String type, String name)
{
  Vector args=new Vector();
  method_header("get_"+name, type, args);
  return_statement(name);
  method_footer();
}

public void set_method(String type, String name)
{
  Vector args=new Vector();
  args.addElement(new Pair(type, "new"+name));
  method_header("set_"+name, "void", args);
  assignment_statement(name, "new"+name);
  method_footer();
}

abstract public void method_header(String name, String returntype, Vector args);

abstract public void return_statement(String to_return);

abstract public void assignment_statement(String to, String from);

public void method_body(Vector source)
{
        Enumeration it=source.elements();
        while(it.hasMoreElements())
          out.println(it.nextElement());
}

abstract public void method_footer();

public void create_method(String name, Property prop)
{
      if(prop!=null)
      {
      if(prop.type.equals("Function"))
      {
        Function f=(Function)prop.object;
        if(f.get_returnType().equals(pc.name()))
        { 
         Vector ar=f.get_args();
         if(ar.size()==1)
         {
          p=(Pair)ar.elementAt(0);
          if(p.first.equals("String"))
          {
           method_header("create", name, ar);
           method_body(f.get_sourceCode());
           method_footer();
          }
         }
        }
      }
     }
}

public void other_methods(Hashtable properties)
{
      iterator=properties.keys();
      while(iterator.hasMoreElements())
      {
        Enumeration it;

        s=(String)iterator.nextElement();
        System.out.println("Checking "+s);
        if(s.equals("create")) continue;
        prop=(Property)pc.getproperties().get(s);
        if(!prop.type.equals("Function")) continue;
        Function f=(Function)prop.object;
        method_header("exec"+s, Database.main.types.classname(f.get_returnType()), f.get_args());
        System.out.println("Contains "+f.get_sourceCode().size()+" lines");
        method_body(f.get_sourceCode());
        method_footer();
      }
}

abstract public void class_footer();

  public void preprocess(String name) throws InterruptedException, IOException
  {
    Process proc;

    System.out.println("Preprocessing...");
  }

  public void process(String name) throws InterruptedException, IOException
  {
    Process proc;
    char c;
    InputStream in;

    proc=rt.exec("compile "+name);
    System.out.println("Compiling...");
    in=proc.getInputStream();
    proc.waitFor();
/*    do
    {
      c=(char)in.read();
      if(c!=-1) System.out.write(c);
    }
    while(c!=-1);
*/    in.close();
  }

public void cleanup(String name)
{
  File f=new File(name+"."+file_extension);
  f.renameTo(new File("src/"+name+"."+file_extension));
  f=new File(name+".class");
  f.renameTo(new File("../../user/mud/user/"+name+".class"));
}

public String print(Property prop)
{
  String type=prop.type;
  Object obj=prop.object;

  if(type.equals("Integer"))
  {
    return "new Integer("+obj+")";
  }
  else if(Types.list.isInstance(obj))
  {
    return "new MudArray(\""+obj+"\")";
  }
  else if(type.equals("String"))
  {
    return "\""+obj+"\"";
  }
  else if(Types.object.isInstance(obj))
  {
    MObject mo=(MObject)obj;
    String p="";
    p=mo.getparent().name();
    if(Database.types.checktype(p, obj))
    {
      MudObject mudobj=(MudObject)obj;
      return mudobj.toString();
    }
    else
      return "null";
  }
  return "null";
}
}
