package momoko.extra;

import momoko.*;
import momoko.user.*;
import java.util.*;
import org.thalassocracy.irc.*;
import java.io.IOException;

public class GenericIrcRoom extends GenericRoom implements IRCRoom, IrcEventListener
{
  static public boolean debug=false;
  Map userToClient;
  Map userToIrc=new Hashtable();
  Map ircToUser=new Hashtable();
  Map ircPeople=new Hashtable();
  IrcClient client=null;
  IrcChannel roomChannel=null;
  String server;
  Thread t;

  public GenericIrcRoom()
  {
    this("");
  }

  public GenericIrcRoom(String name)
  {
    super(name);
  }

  public GenericIrcRoom(String server, Map clients, String name)
  {
    this(name);
    this.server=server;
    userToClient=clients;
/*
    client=ic;
    client.addEventListener(this);

    try
    {
      roomChannel=new IrcChannel(client.irc, name);
      roomChannel.join();
    }
    catch(Exception e) {e.printStackTrace();}
*/
  }

  public IrcClient getClientForUser(IMObject user)
  {
    return (IrcClient)userToClient.get(user);
  }

  public void handleEvent(IrcEvent foo)
  {
//if(debug) System.out.println("Event! "+foo);
    if(foo instanceof IrcCtcplessPrivmsgEvent)
      handleEvent((IrcCtcplessPrivmsgEvent)foo);
    if(foo instanceof IrcCtcpActionEvent)
      handleEvent((IrcCtcpActionEvent)foo);
    if(foo instanceof IrcPartEvent)
      handleEvent((IrcPartEvent)foo);
    if(foo instanceof IrcJoinEvent)
      handleEvent((IrcJoinEvent)foo);
    if(foo instanceof IrcQuitEvent)
      handleEvent((IrcQuitEvent)foo);
    if(foo instanceof IrcKickEvent)
      handleEvent((IrcKickEvent)foo);
  }

  public void handleEvent(IrcCtcplessPrivmsgEvent foo)
  {
    if (foo.msgtarget.equals(getName()))
    {
      foo.sender=debang(foo.sender);
      IMObject sayer=(IMObject)subdir.get(foo.sender);
      if(sayer==null) announce(foo.sender, foo.sender+" says, \""+foo.message+"\"");
      else announce(sayer, foo.sender+" says, \""+foo.message+"\"");
    }
    else
    {
      IUser user=null;
      IrcUser sender=null;
      try
      {
        user=(IUser)resolve(debang(foo.msgtarget));
        sender=(IrcUser)resolve(debang(foo.sender));
      }
      catch(Exception e) {e.printStackTrace(); return;}
      user.whisper(sender, foo.message);
    }
  }

  public void handleEvent(IrcCtcpActionEvent foo)
  {
    if (foo.msgtarget.equals(getName()))
    {
      foo.sender=debang(foo.sender);
      IMObject sayer=(IMObject)subdir.get(foo.sender);
      if(sayer==null) announce(foo.sender, foo.sender+" "+foo.message);
      else announce(sayer, foo.sender+" "+foo.message);
    }
  }

  public void handleEvent(IrcPartEvent foo)
  {
    foo.sender=debang(foo.sender);
    if(foo.channel.equals(getName()))
    {
      tellAllBut(foo.sender, foo.sender+" goes to get some more tea.");
    }
  }

  public void handleEvent(IrcQuitEvent foo)
  {
    foo.sender=debang(foo.sender);
    tellAllBut(foo.sender, foo.sender+" suddenly falls through the floor.");
  }

  public void handleEvent(IrcKickEvent foo)
  {
    foo.sender=debang(foo.sender);
    if(foo.channel.equals(getName()))
    {
      tellAllBut(foo.sender, foo.sender+" is consumed by a giant carnivorous plant.");
    }
  }

  public void handleEvent(IrcJoinEvent foo)
  {
    foo.sender=debang(foo.sender);
    if(foo.channel.equals(getName()))
      tellAllBut(foo.sender, foo.sender+" meanders in with a steaming cup of Earl Grey.");
  }

  private String debang(String s)
  {
    int i=s.indexOf('!');
    if(i==-1) return s;
    else return s.substring(0, i);
  }

  public void enterfunc(IMObject object)
  {
if(debug) System.out.println("enterfunc: "+object.getName());
    IrcClient newclient=(IrcClient)userToClient.get(object);
    if(newclient==null)
    {
      if(debug) System.out.println("new client for "+object.getName()+"!");
      try
      {
        System.out.println("location: "+location());
        newclient=new IrcClient(server, 6667, object.getName(), object.getName(), object.getName());
        userToClient.put(object, newclient);
      }
      catch(Exception e) {e.printStackTrace();}
    }

    if(client==null)
    {
      client=newclient;
      client.addEventListener(this);
    }

    IrcChannel newchannel=(IrcChannel)userToIrc.get(object);
    if(newchannel==null)
    {
      if(debug) System.out.println("new channel for "+object.getName()+"!");
      newchannel=new IrcChannel(newclient.irc, getName());
      try
      {
        newchannel.join();
        userToIrc.put(object, newchannel);
        ircToUser.put(newchannel, object);
      }
      catch(Exception e) {e.printStackTrace();}
    }
    if(roomChannel==null)
    {
      roomChannel=newchannel;
    }
if(debug) System.out.println("Entered");
  }

public void exitfunc(IMObject obj)
{
   IrcChannel channel=(IrcChannel)userToIrc.get(obj);
   if(channel!=null)
   {
     try { channel.part(); } catch(Exception e) {e.printStackTrace();}
     ircToUser.remove(channel);
   }
   userToIrc.remove(obj);
   if(channel.equals(roomChannel))
   {
     Iterator iterator=userToIrc.keySet().iterator();
     if(!iterator.hasNext())
     {
       roomChannel=null;
     }
     else
     {
       obj=(IMObject)iterator.next();
       roomChannel=(IrcChannel)userToIrc.get(obj);
     }
   }

   if(client==null) return;
   IrcClient cl=(IrcClient)userToClient.get(obj);
   if(cl==null) return;
   if(cl.equals(client))
   {
     cl.removeEventListener(this);
     Iterator iterator=userToClient.keySet().iterator();
     if(iterator.hasNext())
     {
       obj=(IMObject)iterator.next();
       client=(IrcClient)userToClient.get(obj);
       if(client!=null)
         client.addEventListener(this);
     }
     else
     {
       client=null;
     }
   }
}

  public void say(IMObject player, String args)
  {
    try
    {
      IUser user=(IUser)player;
      user.tell("You say, \""+args+"\"");
      IrcChannel channel=(IrcChannel)userToIrc.get(user);
      if(channel==null)
        System.out.println("User not on a channel!");
      else
        try { channel.privmsg(args); }
        catch(Exception e) {e.printStackTrace();}
    }
    catch(ClassCastException e) {if(debug) System.out.println("Sayer is not a user!");}
  }

  public void emote(IMObject player, String args)
  {
    try
    {
      IUser user=(IUser)player;
      user.tell(user.name()+" "+args);
      IrcChannel channel=(IrcChannel)userToIrc.get(user);
      if(channel==null)
        System.out.println("User not on a channel!");
      else
        try { channel.ctcpAction(args); }
        catch(Exception e) {e.printStackTrace();}
    }
    catch(ClassCastException e) {if(debug) System.out.println("Sayer is not a user!");}
  }

  public void announce(IMObject player, String args)
  {
    Iterator iterator=subdir.keySet().iterator();
    while(iterator.hasNext())
    {
      String name=(String)iterator.next();
      MObject mo=(MObject)subdir.get(name);
      IUser user=null;
      try
      {
        user=(IUser)mo;
        if(user==player) continue;
        user.tell(args);
      }
      catch(ClassCastException e) {}
    }
  }

  public void announce(String player, String args)
  {
    Iterator iterator=subdir.keySet().iterator();
    while(iterator.hasNext())
    {
      String name=(String)iterator.next();
      MObject mo=(MObject)subdir.get(name);
      IUser user=null;
      try
      {
        user=(IUser)mo;
        if(user.getName()!=player)
        user.tell(args);
      }
      catch(Exception e) {}
    }
  }

  public void tellAllBut(String player, String args)
  {
    Iterator iterator=subdir.keySet().iterator();
    while(iterator.hasNext())
    {
      String name=(String)iterator.next();
      MObject mo=(MObject)subdir.get(name);
      IUser user=null;
      try
      {
        user=(IUser)mo;
        if(!user.getName().equals(player))
          user.tell(args);
      }
      catch(Exception e) {}
    }
  }

  public Map contents()
  {
    Vector v=null;
//    if(debug) System.out.println(roomChannel);
//    if(debug) System.out.println(roomChannel.getName());

    Map h=subdir;

    if(roomChannel!=null)
    {
      try { v=roomChannel.getUsers(); }
      catch(Exception e) {e.printStackTrace();}

      Enumeration iterator=v.elements();
      while(iterator.hasMoreElements())
      {
        String s=(String)iterator.nextElement();
        if(h.get(s)==null)
          h.put(s, new IrcUser(s));
      }
    }

    return h;
  }
}
