package momoko.forum;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;
import momoko.*;
import java.lang.reflect.*;
import javax.mail.*;
import momoko.proxy.*;
import momoko.protoclass.*;

/**
 *  a multi-user aware object
 */
public class Board extends MObject
{
  public static boolean debug=false;
  transient boolean built=false;

/**
 *  create a new object without a name
 */
public Board()
{
  this("");
}

/**
 *  create a new object with the given name
 */
public Board(String newname)
{
  super(newname);
  setcontents(new TreeMap(new StringIntComparator()));
}

public IMObject create(IProtoClass parent, String path, String name) throws IllegalArgumentException, BadPathException
{
  return super.create(parent, path, name);
}

private IMObject regenerate(String name)
{
  if(debug) System.out.println("registering server "+name);
  IMObject server=null;
  try
  {
    if(subdir.size()==0)
      buildThreads();
    server=(IMObject)subdir.get(name);
  }
  catch(Exception e) {e.printStackTrace(); return null;}
  return server;
}

public Map contents()
{
  if(!built)
    buildThreads();
  return super.contents();
}

public IMObject resolve(String path) throws BadPathException
{
  String prefix, suffix, temp=null;
  int i;
  boolean done=false;
  Enumeration iterator;
 
  i=path.indexOf('/');
  if(i==-1)
  {
    prefix=path;
    suffix="";
  }
  else
  {
    prefix = path.substring(0, i);
    suffix = path.substring(i+1);
  }

  IMObject server=regenerate(prefix);
  if(server==null) throw new BadPathException(path);

  if(suffix.equals(""))
  {
    return server;
  }
  else
    return server.resolve(suffix);
}

public void buildThreads()
{
  Hashtable id_table = new Hashtable();
  built=true;
  try
  {
    MailFolder folder=(MailFolder)Database.main.root.resolve("/world/mail/mh/"+getName());
    IFolder board=folder.getFolder();
    board.open(Folder.READ_WRITE);
    if(debug) System.out.println("board: "+board);
    IMessage[] messages=board.getMessages();
    if(debug) System.out.println("messages: "+messages);
    if(debug) System.out.println("count: "+messages.length);
    String id=null;
    for(int x=0;x<messages.length;x++)
    {
      id=new Integer(messages[x].getMessageNumber()).toString();
      if(debug) System.out.println(x+": "+messages[x]);
      IMObject m=(IMObject)MObjectProxy.newInstance(id, messages[x]);
      m.setcontents(new TreeMap(new StringIntComparator()));
      if(debug) System.out.println("m: "+m);
      id_table.put(id, m);
    }

    if(debug) System.out.println("Made message containers");

    for(int x=0;x<messages.length;x++)
    {
      id=new Integer(messages[x].getMessageNumber()).toString();
      IMessage m=(IMessage)id_table.get(id);
      IMObject container=null;
      try
      {
        id=getId(m);
        container=(IMObject)id_table.get(id);
      }
      catch(MessagingException e) {}
      if(container==null)
        container=this;
      if(debug) System.out.println("board.move "+m+" "+container);
      Database.main.move((IMObject)m, container);
    }

    if(debug) System.out.println("Established message relationships");
  }
  catch(Exception e) {e.printStackTrace();}
  finally {built=true;}
}

public String getId(IMessage m) throws MessagingException
{
  String id=null;
  String[] headers;

  headers=m.getHeader("References");
  // I don't yet handle References
  headers=m.getHeader("In-Reply-To");
  if(headers!=null)
    if(headers.length>0)
      id=headers[0];
  if(id!=null) return id;
  else throw new MessagingException();
}

private class StringIntComparator implements Comparator
{
  public int compare(Object o1, Object o2)
  {
    try
    {
      String s1=(String)o1;
      String s2=(String)o2;
      Integer i1=new Integer(s1);
      Integer i2=new Integer(s2);
      return i1.compareTo(i2);
    }
    catch(Exception e) { return 0; }
  }
}

}
