package momoko.protoclass;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;
import java.io.*;
import momoko.*;
import momoko.server.Module;
import java.lang.reflect.*;
import java.net.*;
import momoko.webserver.*;
import momoko.versioncontrol.*;

/**
 *  a multi-user aware object
 */
public class GenericClassRepository extends MObject implements ClassRepository
{
  private AdaptiveClassLoader loader;
  protected static boolean debug = true;

/**
 *  create a new object without a name
 */
public GenericClassRepository()
{
  this("");
}

/**
 *  create a new object with the given name
 */
public GenericClassRepository(String newname)
{
  super(newname);
  loader=(AdaptiveClassLoader)Database.main.loader;
}

public void setClassLoader(AdaptiveClassLoader l)
{
  loader=l;
}

public AdaptiveClassLoader getClassLoader()
{
  return loader;
}

public void initialize(Properties args)
{
}

public IMObject create(IProtoClass parent, String path, String name) throws IllegalArgumentException, BadPathException
{
  return super.create(parent, path, name);
}

private IMObject regenerate(String name)
{
  if(debug) System.out.println("registering class "+name);
  if(debug) System.out.println("ClassRepository: "+this);
  VersionController vc=null;
  try
  {
    IMObject obj=(IMObject)subdir.get(name);
    if(obj!=null) return obj;

if(debug) System.out.println("loader: "+loader);
      Class c=loader.loadClass(name);
if(debug) System.out.println("c: "+c);
      vc=new GenericVersionController(name);
      Database.main.move(vc, this);
      IProtoClass pc=vc.getCurrent();
      pc.initialize("momoko.MObject");
      pc.setClassLoader(loader);
      Field[] f=c.getDeclaredFields();
      int i=0;
//      if(debug) System.out.println(name+" has "+f.length+" declared fields.");
      while(i<f.length)
      {
//        if(debug) System.out.println("```:"+f[i].getName()+" "+f[i].getType().getName());
        pc.addprop(f[i].getName(), f[i].getType().getName(), null);
        i=i+1;
      }
//      if(debug) System.out.println("done with fields");
  }
  catch(ClassNotFoundException e)
  {
if(debug) System.out.println("url checking");
    URL url=loader.getResource(name);
if(debug) System.out.println("url: "+url);
    if(url==null) return null;
    else return new GenericResource(url);
  }
  catch(Exception e) {e.printStackTrace(); return null;}
  return vc;
}

public IMObject resolve(String path) throws BadPathException
{
if(debug) System.out.println("Repository looking for "+path); 
  IMObject o=regenerate(path);
  if(o==null)
  {
    System.out.println("Class not found");
    throw new BadPathException(path);
  }
  return o;
}

}
