package momoko.relationship;
import java.util.*;
import java.io.Serializable;
import momoko.*;
import momoko.types.*;

public class SubjectExclusiveRelationship implements Relationship,
Serializable
{
  static public boolean debug=false;
  boolean local;
  String name;
  Map subjectkeyed=new HashMap();
  Map objectkeyed=new HashMap();
  Predicate postrelation=null;
  Predicate postunrelation=null;

  public SubjectExclusiveRelationship(String newname, boolean locality)
  {
    name=newname;
    local=locality;
  }

  public void print(IMObject mobj)
  {
    Iterator iterator=objectkeyed.keySet().iterator();
    while(iterator.hasNext())
    {
      IMObject o=(IMObject)iterator.next();
      if(o.getName().equals(mobj.getName()))
        System.out.println("o: "+o+" -- "+o.getName());
    }
  }

  public void setfunctions(Predicate postr, Predicate postu)
  {
    postrelation=postr;
    postunrelation=postu;
  }

  public String name()
  {
    return name;
  }

  public void relate(Object subject, Object object) throws ExclusivityException
  {
    Object o;
    Vector temp;

    if(debug) System.out.println(subject+" "+object);
    if(debug) System.out.println("** Relating "+((IMObject)subject).name()+" to "+((IMObject)object).name());
    o=objectkeyed.get(object);

    if(o==subject)
    {
      return;
    }

    if(o!=null) 
    {
      if(debug) System.out.println("Relationship database already contains "+o+" "+((IMObject)o).getName());
      throw new ExclusivityException();
    }

    Object tobj=subjectkeyed.get(subject);
    temp=(Vector)tobj;
//    temp=(Vector)subjectkeyed.get(subject);
    if(temp==null)
      temp=new Vector();
    temp.addElement(object);
    subjectkeyed.put(subject, temp);
    Object confirm=subjectkeyed.get(subject);
    if(confirm==null) throw new NullPointerException("Subject table corrupted");
    if(debug) System.out.println("object put "+object+" "+subject);
    if(debug) System.out.println("--PUT");
    objectkeyed.put(object, subject);
    if(debug) System.out.println("PUT--");
    if(debug) System.out.println("--GET");
    confirm=objectkeyed.get(object);
    if(debug) System.out.println("GET--");
    if(confirm==null)
    {
      if(debug) System.out.println("Unsupported object: "+object);
      throw new NullPointerException("Object table corrupted");
    }
    if(debug) System.out.println("Confirm: "+objectkeyed.get(object));
    if(postrelation!=null)
      postrelation.execute(subject, object);
    if(Types.object.isInstance(subject))
      if(!local|((IMObject)subject).isLocal())
      {
        if(debug) System.out.println("** Storing "+name);
        ((IMObject)subject).getrelationships().put(name, this);
      }
    if(Types.object.isInstance(object))
      if(!local|((IMObject)object).isLocal())
        ((IMObject)object).getrelationships().put(name, this);
  }

  public void relateforced(Object subject, Object object)
  {
    try { relate(subject, object); }
    catch(ExclusivityException e)
    { 
      if(debug) System.out.println("Exclusivity conflict. Resolving...");
      unrelate(subject(object), object);
      try { relate(subject, object); }
      catch(ExclusivityException f)
      {
        System.out.println("The relationship database is weird. relateforced().");
      }
    }
  }

  public Object subject(Object object)
  {
    //print((IMObject)object);
    return objectkeyed.get(object);
  }

  public Object object(Object subject)
  {
    Vector temp;

    temp=(Vector)subjectkeyed.get(subject);
    if(temp==null)
      return null;
    else
      return temp.elementAt(0);
  }

  public Vector subjects(Object object)
  {
    Object o;
    Vector temp;

    o=objectkeyed.get(object);
    if(o==null)
      return null;
    else
    {
      temp=new Vector();
      temp.addElement(o);
      return temp;
    }
  }

  public Vector objects(Object subject)
  {
    return (Vector)subjectkeyed.get(subject);
  }

  public void unrelate(Object subject, Object object)
  {
    Object o;
    Vector temp;

    o=objectkeyed.get(object);
    if(o!=subject) return;
      objectkeyed.remove(object);
    temp=(Vector)subjectkeyed.get(subject);
    if(temp==null)
    {
//      System.out.println("No relationships for "+subject);
    }
    else
    {
      temp.remove(object);
      subjectkeyed.put(subject, temp);
    }
    objectkeyed.remove(object);
    if(postunrelation!=null)
      postunrelation.execute(subject, object);
    if(Types.object.isInstance(subject))
      if(!local|((MObject)subject).isLocal())
        ((MObject)subject).getrelationships().remove(name);
    if(Types.object.isInstance(object))
      if(!local|((MObject)object).isLocal())
        ((MObject)object).getrelationships().remove(name);
  }

  public void remove(Object thing)
  {
  }
}
