package org.thalassocracy.irc;

import java.net.*;
import java.io.*;
import java.util.*;
import java.beans.*;

public class BaseIrcClient implements IrcEventDispatcher, IrcEventListener {

    final static String MQUOTE = "\020";
    final static String XQUOTE = "\\";
    final static String XDELIM = "\001";

    Socket socket;
    BufferedReader in;
    BufferedWriter out;

    String host = "localhost";
    int port = 6669;


    public BaseIrcClient ()
    {
    }

    public BaseIrcClient (String host, int port)
    {

        /* should I really be handling my own events?
           or should this go in the trigger methods
           or in the code around the trigger calls? */

        this.addEventListener(this);

        try {
            this.socket = new Socket(host, port);
            this.in = new BufferedReader
                (new InputStreamReader(this.socket.getInputStream()));
            this.out = new BufferedWriter
                (new OutputStreamWriter(this.socket.getOutputStream()));
        } catch (IOException e) {
            e.printStackTrace();
        }

        this.host = host;
        this.port = port;
    }


    private Vector eventListeners = new Vector();
    public void addEventListener (IrcEventListener listener)
    {
        eventListeners.addElement(listener);
    }
    public void removeEventListener (IrcEventListener listener)
    {
        eventListeners.removeElement(listener);
    }

    private void triggerEvent (IrcEvent foo)
    {
        for (Enumeration e = eventListeners.elements(); e.hasMoreElements();) {
            ((IrcEventListener)e.nextElement()).handleEvent(foo);
        }
    }

    public String getHost ()
    {
        return host;
    }

    public int getPort ()
    {
        return port;
    }

    final private void send(String msg) throws IrcMessageError
    {
        if ((msg.indexOf('\000') != -1)
            || (msg.indexOf('\r') != -1)
            || (msg.indexOf('\n') != -1)) {
            throw new IrcMessageError();
        }
        try {
            this.out.write(msg + "\n");
            this.out.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }


    private String lowQuote(String msg)
    {
        msg = StringUtils.replace(msg, MQUOTE, MQUOTE + MQUOTE);
        msg = StringUtils.replace(msg, "\000", MQUOTE + "0");
        msg = StringUtils.replace(msg, "\r", MQUOTE + "r");
        msg = StringUtils.replace(msg, "\n", MQUOTE + "n");
        return msg;
    }

    private String lowDequote(String msg)
    {
        String dmsg = "";
        char[] omsg;
        int i;

        omsg = msg.toCharArray();

        i = 0;
        while (i<msg.length()) {
            if (String.valueOf(omsg[i]).equals(MQUOTE)) {
                if (String.valueOf(omsg[i+1]).equals(MQUOTE)) {
                    dmsg = dmsg + MQUOTE;
                    i = i + 2;
                } else if (omsg[i+1] == 'n') {
                    dmsg = dmsg + "\n";
                    i = i + 2;
                } else if (omsg[i+1] == 'r') {
                    dmsg = dmsg + "\r";
                    i = i + 2;
                } else if (omsg[i+1] == '0') {
                    dmsg = dmsg + "\000";
                    i = i + 2;
                } else {
                    i = i + 1;
                }
            } else {
                dmsg = dmsg + omsg[i];
                i = i + 1;
            }
        }

        return dmsg;

    }

    private String ctcpQuote(String msg)
    {
        msg = StringUtils.replace(msg, XQUOTE, XQUOTE +  XQUOTE);
        msg = StringUtils.replace(msg, XDELIM, XQUOTE + "a");
        msg = this.lowQuote(msg);
        return msg;
    }

    private String ctcpDequote(String msg)
    {
        String dmsg = "";
        char[] omsg;
        int i;

        msg = this.lowDequote(msg);

        omsg = msg.toCharArray();

        i = 0;
        while (i<msg.length()) {
            if (String.valueOf(omsg[i]).equals(XQUOTE)) {
                if (String.valueOf(omsg[i+1]).equals(XQUOTE)) {
                    dmsg = dmsg + XQUOTE;
                    i = i + 2;
                } else if (omsg[i+1] == 'a') {
                    dmsg = dmsg + XDELIM;
                    i = i + 2;
                } else {
                    i = i + 1;
                }
            } else {
                dmsg = dmsg + omsg[i];
                i = i + 1;
            }
        }

        return dmsg;

    }


    public void pass(String password) throws IrcMessageError
    {
        this.send("pass " + password);
    }

    public void nick(String name) throws IrcMessageError
    {
        this.send("nick " + name);
    }

    public void user(String username, String hostname, String realname)
        throws IrcMessageError
    {
        this.send("user " + username + " " + hostname + " " + this.host +
                  " :" + realname);
    }

    public void oper(String username, String password) throws IrcMessageError
    {
        this.send("oper " + username + " " + password);
    }

    public void quit(String msg) throws IrcMessageError
    {
        this.send("quit :" + msg);
    }

    public void quit() throws IrcMessageError
    {
        this.send("quit");
    }

    public void join(String channel, String key) throws IrcMessageError
    {
        this.send("join " + channel + " " + key);
    }

    public void join(String channel) throws IrcMessageError
    {
        this.send("join " + channel);
    }

    public void part(String channel) throws IrcMessageError
    {
        this.send("part " + channel);
    }

    public void part(String channel, String message) throws IrcMessageError
    {
        this.send("part " + channel + " :" + message);
    }

    // needs work
    public void chanmode(String channel, String modestr) throws IrcMessageError
    {
        this.send("mode " + channel + " " + modestr);
    }

    // needs work
    public void usermode(String user, String modestr) throws IrcMessageError
    {
        this.send("mode " + user + " " + modestr);
    }

    public void topic(String channel, String topic) throws IrcMessageError
    {
        this.send("topic " + channel + " :" + topic);
    }

    public void names(String channel) throws IrcMessageError
    {
        this.send("names " + channel);
    }

    // figure out what to do with the optional "server" parameter
    public void list(String channels) throws IrcMessageError
    {
        this.send("list " + channels);
    }

    public void list() throws IrcMessageError
    {
        this.send("list");
    }

    public void invite(String nickname, String channel) throws IrcMessageError
    {
        this.send("invite " + nickname + " " + channel);
    }

    public void kick(String channel, String user, String comment)
        throws IrcMessageError
    {
        this.send("kick " + channel + " " + user + " :" + comment);
    }

    public void kick(String channel, String user) throws IrcMessageError
    {
        this.send("kick " + channel + " " + user);
    }

    public void version(String server) throws IrcMessageError
    {
        this.send("version " + server);
    }

    public void version() throws IrcMessageError
    {
        this.send("version");
    }

    // needs work
    public void stats(String query, String server) throws IrcMessageError
    {
        this.send("stats " + query + " " + server);
    }

    public void stats(String query) throws IrcMessageError
    {
        this.send("stats " + query);
    }

    public void links(String remote, String mask) throws IrcMessageError
    {
        this.send("links " + remote + " " + mask);
    }

    public void links(String remote) throws IrcMessageError
    {
        this.send("links " + remote);
    }

    public void links() throws IrcMessageError
    {
        this.send("links");
    }

    public void time(String server) throws IrcMessageError
    {
        this.send("time " + server);
    }

    public void time() throws IrcMessageError
    {
        this.send("time");
    }

    public void connect(String server, String port, String remote)
        throws IrcMessageError
    {
        this.send("connect " + server + " " + port + " " + remote);
    }

    public void connect(String server, String port) throws IrcMessageError
    {
        this.send("connect " + server + " " + port);
    }

    // do something about that default port
    public void connect(String server) throws IrcMessageError
    {
        this.send("connect " + server + " 6669");
    }

    public void trace(String server) throws IrcMessageError
    {
        this.send("trace " + server);
    }

    public void trace() throws IrcMessageError
    {
        this.send("trace");
    }

    public void admin(String server) throws IrcMessageError
    {
        this.send("admin " + server);
    }

    public void admin() throws IrcMessageError
    {
        this.send("admin");
    }

    public void info(String server) throws IrcMessageError
    {
        this.send("info " + server);
    }

    public void info() throws IrcMessageError
    {
        this.send("info");
    }

    public void privmsg(String receiver, String text) throws IrcMessageError
    {
        this.send("privmsg " + receiver + " :" + text);
    }

    public void notice(String user, String text) throws IrcMessageError
    {
        this.send("notice " + user + " :" + text);
    }

    public void who(String name) throws IrcMessageError
    {
        this.send("who " + name);
    }

    public void who() throws IrcMessageError
    {
        this.send("who");
    }

    public void whoops(String name) throws IrcMessageError
    {
        this.send("who " + name + " o");
    }

    public void whoops() throws IrcMessageError
    {
        this.send("who " + "0" + " o");
    }

    public void whois(String nickmask, String server) throws IrcMessageError
    {
        this.send("whois " + server + " " + nickmask);
    }

    public void whois(String nickmask) throws IrcMessageError
    {
        this.send("whois " + nickmask);
    }

    public void whowas(String nickname, String count, String server)
        throws IrcMessageError
    {
        server.trim();
        count.trim();
        if ((!server.equals("")) && (count.equals(""))) {
            count = "-1";
        }
        this.send("whowas " + nickname + " " + count + " " + server);
    }

    public void whowas(String nickname, String count) throws IrcMessageError
    {
        this.send("whowas " + nickname + " " + count);
    }

    public void whowas(String nickname) throws IrcMessageError
    {
        this.send("whowas " + nickname);
    }

    public void kill(String nickname, String comment) throws IrcMessageError
    {
        this.send("kill " + nickname + " " + comment);
    }

    public void ping(String server) throws IrcMessageError
    {
        this.send("ping " + server);
    }

    public void pong(String daemon) throws IrcMessageError
    {
        this.send("pong " + daemon);
    }

    public void away(String message) throws IrcMessageError
    {
        this.send("away :" + message);
    }

    public void unaway() throws IrcMessageError
    {
        this.send("away");
    }

    public void rehash() throws IrcMessageError
    {
        this.send("rehash");
    }

    public void restart() throws IrcMessageError
    {
        this.send("restart");
    }

    public void summon(String user, String server) throws IrcMessageError
    {
        this.send("summon " + user + " " + server);
    }

    public void summon(String user) throws IrcMessageError
    {
        this.send("summon " + user);
    }

    public void users(String server) throws IrcMessageError
    {
        this.send("users " + server);
    }

    public void users() throws IrcMessageError
    {
        this.send("users");
    }

    public void userhost(String nickname) throws IrcMessageError
    {
        this.send("userhost " + nickname);
    }

    public void ison(String nickname) throws IrcMessageError
    {
        this.send("ison " + nickname);
    }

    public void handleEvent (IrcEvent foo)
    {
        if (foo instanceof IrcNoticeEvent) {
            this.handleEvent((IrcNoticeEvent) foo);
        } else if (foo instanceof IrcPrivmsgEvent) {
            this.handleEvent((IrcPrivmsgEvent) foo);
        }
    }

    public void handleEvent (IrcPrivmsgEvent privmsg)
    {
        int i;
        String msg;
        String ctcpmsg;
        String cmd;
        String sender;
        String msgtarget;
        String message;
        StringTokenExtractor c;
        StringTokenExtractor d;

        sender = privmsg.sender;
        msgtarget = privmsg.msgtarget;
        message = privmsg.message;

        c = new StringTokenExtractor(message, XDELIM);

        i = 0;
        msg = "";
        while (c.hasMoreTokens()) {
            msg = msg + c.nextToken();
            if (c.hasMoreTokens()) {
                ctcpmsg=ctcpDequote(c.nextToken());
                d = new StringTokenExtractor(ctcpmsg, " ");
                cmd = d.nextToken();
                ctcpmsg = d.getRemainder();
                if (cmd.equals("ACTION")) {
                    IrcCtcpActionEvent foo = new IrcCtcpActionEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    foo.message = ctcpmsg;
                    this.triggerEvent(foo);
                } else if (cmd.equals("DCC")) {
                    /* FIXME

                       this should be more specific and throw one of
                       the real Dcc(Foo)Events.  Or maybe I should
                       have a handler do that.  Or maybe the DccEvent
                       trigger should do that.

                    */

                    IrcDccEvent foo = new IrcDccEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("SED")) {
                    IrcCtcpSedEvent foo = new IrcCtcpSedEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("FINGER")) {
                    IrcCtcpFingerEvent foo = new IrcCtcpFingerEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("VERSION")) {
                    IrcCtcpVersionEvent foo = new IrcCtcpVersionEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("SOURCE")) {
                    IrcCtcpSourceEvent foo = new IrcCtcpSourceEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("USERINFO")) {
                    IrcCtcpUserinfoEvent foo = new IrcCtcpUserinfoEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("CLIENTINFO")) {
                    IrcCtcpClientinfoEvent foo = new IrcCtcpClientinfoEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else if (cmd.equals("ERRMSG")) {
                    IrcCtcpErrmsgEvent foo = new IrcCtcpErrmsgEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    foo.message = ctcpmsg;
                    this.triggerEvent(foo);
                } else if (cmd.equals("PING")) {
                    IrcCtcpPingEvent foo = new IrcCtcpPingEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    foo.timestamp = ctcpmsg;
                    this.triggerEvent(foo);
                } else if (cmd.equals("TIME")) {
                    IrcCtcpTimeEvent foo = new IrcCtcpTimeEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                } else {
                    IrcCtcpEvent foo = new IrcCtcpEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                }
            }
        }

        if (!msg.equals("")) {
            IrcCtcplessPrivmsgEvent foo = new IrcCtcplessPrivmsgEvent();
            foo.name = privmsg.name;
            foo.sender = sender;
            foo.args = msg;
            foo.msgtarget = msgtarget;
            foo.message = msg;
            this.triggerEvent(foo);
        }

    }

    public void handleEvent (IrcNoticeEvent notice)
    {
        int i;
        String msg;
        String ctcpmsg;
        String cmd;
        String sender;
        String msgtarget;
        String message;
        StringTokenExtractor c;
        StringTokenExtractor d;

        sender = notice.sender;
        msgtarget = notice.msgtarget;
        message = notice.message;

        c = new StringTokenExtractor(message, XDELIM);

        i = 0;
        msg = "";
        while (c.hasMoreTokens()) {
            msg = msg + c.nextToken();
            if (c.hasMoreTokens()) {
                ctcpmsg=ctcpDequote(c.nextToken());
                d = new StringTokenExtractor(ctcpmsg, " ");
                cmd = d.nextToken();
                ctcpmsg = d.getRemainder();
                if (cmd.equals("FINGER")) {
                    IrcCtcpFingerRplEvent foo = new IrcCtcpFingerRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    try {
                        if (ctcpmsg.charAt(0) == ':') {
                            foo.message = ctcpmsg.substring(1);
                        } else {
                            foo.message = ctcpmsg;
                        }
                    } catch (StringIndexOutOfBoundsException e) {
                        foo.message = "";
                    }
                    this.triggerEvent(foo);
                } else if (cmd.equals("VERSION")) {
                    StringTokenExtractor fields;
                    IrcCtcpVersionRplEvent foo = new IrcCtcpVersionRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    fields = new StringTokenExtractor(ctcpmsg, ":");
                    foo.clientName = fields.nextToken();
                    foo.clientVersion = fields.nextToken();
                    foo.clientEnvironment = fields.nextToken();
                    this.triggerEvent(foo);
                } else if (cmd.equals("SOURCE")) {
                    if (ctcpmsg.length() > 0) {
                        StringTokenExtractor fields;
                        IrcCtcpSourceRplEvent foo =
                            new IrcCtcpSourceRplEvent();
                        foo.name = cmd;
                        foo.sender = sender;
                        foo.args = ctcpmsg;
                        foo.msgtarget = msgtarget;
                        fields = new StringTokenExtractor(ctcpmsg, ":");
                        foo.clientHost = fields.nextToken();
                        foo.clientDirectory = fields.nextToken();
                        foo.clientFiles = fields.nextToken();
                        this.triggerEvent(foo);
                    } else {
                        IrcCtcpSourceFinishedRplEvent foo =
                            new IrcCtcpSourceFinishedRplEvent();
                        foo.name = cmd;
                        foo.sender = sender;
                        foo.args = ctcpmsg;
                        foo.msgtarget = msgtarget;
                        this.triggerEvent(foo);
                    }
                } else if (cmd.equals("USERINFO")) {
                    IrcCtcpUserinfoRplEvent foo =
                        new IrcCtcpUserinfoRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    try {
                        if (ctcpmsg.charAt(0) == ':') {
                            foo.message = ctcpmsg.substring(1);
                        } else {
                            foo.message = ctcpmsg;
                        }
                    } catch (StringIndexOutOfBoundsException e) {
                        foo.message = "";
                    }
                    this.triggerEvent(foo);
                } else if (cmd.equals("CLIENTINFO")) {
                    IrcCtcpClientinfoRplEvent foo =
                        new IrcCtcpClientinfoRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    try {
                        if (ctcpmsg.charAt(0) == ':') {
                            foo.message = ctcpmsg.substring(1);
                        } else {
                            foo.message = ctcpmsg;
                        }
                    } catch (StringIndexOutOfBoundsException e) {
                        foo.message = "";
                    }
                    this.triggerEvent(foo);
                } else if (cmd.equals("ERRMSG")) {
                    StringTokenExtractor fields;
                    IrcCtcpErrmsgRplEvent foo = new IrcCtcpErrmsgRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    fields = new StringTokenExtractor(ctcpmsg, ":");
                    foo.query = fields.nextToken();
                    foo.message = fields.getRemainder();
                    this.triggerEvent(foo);
                } else if (cmd.equals("PING")) {
                    IrcCtcpPingRplEvent foo = new IrcCtcpPingRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    foo.timestamp = ctcpmsg;
                    this.triggerEvent(foo);
                } else if (cmd.equals("TIME")) {
                    IrcCtcpTimeRplEvent foo = new IrcCtcpTimeRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    try {
                        if (ctcpmsg.charAt(0) == ':') {
                            foo.time = ctcpmsg.substring(1);
                        } else {
                            foo.time = ctcpmsg;
                        }
                    } catch (StringIndexOutOfBoundsException e) {
                        foo.time = "";
                    }
                    this.triggerEvent(foo);
                } else {
                    IrcCtcpRplEvent foo = new IrcCtcpRplEvent();
                    foo.name = cmd;
                    foo.sender = sender;
                    foo.args = ctcpmsg;
                    foo.msgtarget = msgtarget;
                    this.triggerEvent(foo);
                }
            }
        }

        if (!msg.equals("")) {
            IrcCtcplessNoticeEvent foo = new IrcCtcplessNoticeEvent();
            foo.name = notice.name;
            foo.sender = sender;
            foo.args = msg;
            foo.msgtarget = msgtarget;
            foo.message = msg;
            this.triggerEvent(foo);
        }
    }



    public void handle(String msg)
    {
        int split;
        String sender;
        String cmd;
        String args;
        int rplnum;
        StringTokenExtractor d;

        if (msg.length() <= 0) {
            return;
        }

        if (msg.charAt(0) == ':') {
            msg = msg.substring(1);
            d = new StringTokenExtractor(msg, " ");
            sender=d.nextToken();
        } else {
            d = new StringTokenExtractor(msg, " ");
            sender = this.host;
        }


        cmd = d.nextToken().toUpperCase();

        msg = d.getRemainder();

        args = msg;

        try {
            rplnum = Integer.valueOf(cmd).intValue();
        } catch (Exception e) {
            rplnum = 0;
        }

        if ((rplnum>=400) && (rplnum<=600)) {
            IrcErrRplEvent foo = new IrcErrRplEvent();
            foo.sender = sender;
            foo.args = args;
            foo.rplnum = rplnum;
            foo.name = new Integer(foo.rplnum).toString();
            foo.message = msg;
            this.triggerEvent(foo);
        } else if (rplnum!=0) {
            IrcRplEvent foo = new IrcRplEvent();
            foo.sender = sender;
            foo.args = args;
            foo.rplnum = rplnum;
            foo.name = new Integer(foo.rplnum).toString();
            foo.message = msg;
            this.triggerEvent(foo);
        } else {
            if (cmd.equals("NICK")) {
                StringTokenExtractor fields;
                IrcNickEvent foo = new IrcNickEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                foo.nickname = fields.nextToken();
                try {
                    if (foo.nickname.charAt(0) == ':') {
                        foo.nickname = foo.nickname.substring(1);
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.nickname = "";
                }
                this.triggerEvent(foo);
            } else if (cmd.equals("QUIT")) {
                IrcQuitEvent foo = new IrcQuitEvent();
                foo.name =cmd;
                foo.sender = sender;
                foo.args = args;
                try {
                    if (msg.charAt(0) == ':') {
                        foo.message = msg.substring(1);
                    } else {
                        foo.message = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.message = "";
                }
                this.triggerEvent(foo);
            } else if (cmd.equals("JOIN")) {
                StringTokenizer channels;
                StringTokenizer keys;
                IrcJoinEvent foo = new IrcJoinEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                try {
                    if (msg.charAt(0) == ':') {
                        msg = msg.substring(1);
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    msg = "";
                }
                channels = new StringTokenizer(msg, ",");
                while (channels.hasMoreTokens()) {
                    foo.channel = channels.nextToken();
                    try {
                        this.triggerEvent((IrcEvent)foo.clone());
                    } catch (CloneNotSupportedException e) {
                        e.printStackTrace();
                    }
                }
            } else if (cmd.equals("PART")) {
                StringTokenExtractor fields;
                StringTokenizer channels;
                IrcPartEvent foo = new IrcPartEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                channels =
                    new StringTokenizer(fields.nextToken(), ",");
                msg = fields.getRemainder();
                try {
                    if (msg.charAt(0) == ':') {
                        foo.message = msg.substring(1);
                    } else {
                        foo.message = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.message = "";
                }
                while (channels.hasMoreTokens()) {
                    foo.channel = channels.nextToken();
                    try {
                        this.triggerEvent((IrcEvent)foo.clone());
                    } catch (CloneNotSupportedException e) {
                        e.printStackTrace();
                    }
                }
            } else if (cmd.equals("MODE")) {

                /* FIXME can a client receive user mode messages as
                   well as channel mode messages?  If so, they should
                   be two separate events */

                /* FIXME furthermore, I should be doing something
                   decent about these modes and modeparams */

                StringTokenExtractor fields;
                IrcModeEvent foo = new IrcModeEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                foo.channel = fields.nextToken();
                foo.modes = fields.nextToken();
                foo.modeparams = fields.nextToken();
                this.triggerEvent(foo);
            } else if (cmd.equals("TOPIC")) {
                StringTokenExtractor fields;
                IrcTopicEvent foo = new IrcTopicEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                foo.channel = fields.nextToken();
                msg = fields.getRemainder();
                try {
                    if (msg.charAt(0) == ':') {
                        foo.message = msg.substring(1);
                    } else {
                        foo.message = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.message = "";
                }
                this.triggerEvent(foo);
            } else if (cmd.equals("INVITE")) {
                StringTokenExtractor fields;
                IrcInviteEvent foo = new IrcInviteEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                fields.nextToken();  /* the server sends me my
                                        own nick.  throw it away.
                                     */
                foo.channel = fields.nextToken();
                this.triggerEvent(foo);
            } else if (cmd.equals("KICK")) {
                StringTokenExtractor fields;
                StringTokenizer channels;
                StringTokenizer nicks;
                IrcKickEvent foo = new IrcKickEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args=args;
                fields = new StringTokenExtractor(msg, " ");
                channels =
                    new StringTokenizer(fields.nextToken(), ",");
                nicks = new StringTokenizer(fields.nextToken(), ",");
                msg = fields.getRemainder();
                try {
                    if (msg.charAt(0) == ':') {
                        foo.message = msg.substring(1);
                    } else {
                        foo.message = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.message = "";
                }
                while (channels.hasMoreTokens()) {
                    foo.channel = channels.nextToken();
                    if (nicks.hasMoreTokens()) {
                        foo.nick = nicks.nextToken();
                    } else {
                        foo.nick = "";
                    }
                    try {
                        this.triggerEvent((IrcEvent)foo.clone());
                    } catch (CloneNotSupportedException e) {
                        e.printStackTrace();
                    }
                }
            } else if (cmd.equals("PRIVMSG")) {
                StringTokenExtractor fields;
                StringTokenizer msgtargets;
                IrcPrivmsgEvent foo = new IrcPrivmsgEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                msgtargets =
                    new StringTokenizer(fields.nextToken(), ",");
                msg = fields.getRemainder();
                try {
                    if (msg.charAt(0) == ':') {
                        foo.message = msg.substring(1);
                    } else {
                        foo.message = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.message = "";
                }
                while (msgtargets.hasMoreTokens()) {
                    foo.msgtarget = msgtargets.nextToken();
                    try {
                        this.triggerEvent((IrcEvent)foo.clone());
                    } catch (CloneNotSupportedException e) {
                        e.printStackTrace();
                    }
                }
            } else if (cmd.equals("NOTICE")) {
                StringTokenExtractor fields;
                StringTokenizer msgtargets;
                IrcNoticeEvent foo = new IrcNoticeEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                fields = new StringTokenExtractor(msg, " ");
                msgtargets =
                    new StringTokenizer(fields.nextToken(), ",");
                msg = fields.getRemainder();
                try {
                    if (msg.charAt(0) == ':') {
                        foo.message = msg.substring(1);
                    } else {
                        foo.message = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.message = "";
                }
                while (msgtargets.hasMoreTokens()) {
                    foo.msgtarget = msgtargets.nextToken();
                    try {
                        this.triggerEvent((IrcEvent)foo.clone());
                    } catch (CloneNotSupportedException e) {
                        e.printStackTrace();
                    }
                }
            } else if (cmd.equals("PING")) {
                IrcPingEvent foo = new IrcPingEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                try {
                    if (msg.charAt(0) == ':') {
                        foo.server = msg.substring(1);
                    } else {
                        foo.server = msg;
                    }
                } catch (StringIndexOutOfBoundsException e) {
                    foo.server = "";
                }
                this.triggerEvent(foo);
            } else {
                IrcEvent foo = new IrcEvent();
                foo.name = cmd;
                foo.sender = sender;
                foo.args = args;
                this.triggerEvent(foo);
            }
        }
    }

    public void handle_cmd () throws IOException, IrcConnectionError
    {
        String msg;

        msg = this.in.readLine();
        if (msg != null) {
            this.handle(msg);
        } else {
            throw new IrcConnectionError();
        }
    }

    public void close() throws IOException
    {
        this.in.close();
        this.out.close();
        this.socket.close();
    }

    public void ctcp(String receiver, String message) throws IrcMessageError
    {
        message = ctcpQuote(message);
        this.privmsg(receiver, XDELIM + message + XDELIM);
    }

    public void ctcpAction(String receiver, String text)
        throws IrcMessageError
    {
        this.ctcp(receiver, "ACTION " + text);
    }

    public void CTCPDCCSend(String receiver, String filename)
    {
        /*dcc_sock = socket.socket(AF_INET, SOCK_STREAM);
          dcc_sock.bind("", 9999);// replace that port # with random (retry)
          dcc_sock.listen(1);
          this.privmsg(receiver, XDELIM + "DCC SEND " + filename + " " +
          address + " " + "9999" + " " + filesize);
          this.await_dccRpl();*/
    }

    public void CTCPDCCChat(String receiver)
    {
        /*dcc_sock = socket.socket(AF_INET, SOCK_STREAM);
          dcc_sock.bind("", 9999);// replace that port # with random (retry)
          dcc_sock.listen(1);
          this.privmsg(receiver, XDELIM + "DCC CHAT chat " + address + " " +
          "9999");
          this.await_dccRpl();*/
    }

    // I don't know how to implement SED encrption -- do you?
    public void ctcpSed()
    {
    }

    public void ctcpFinger(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "FINGER");
    }

    public void ctcpVersion(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "VERSION");
    }

    public void ctcpSource(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "SOURCE");
    }

    public void ctcpUserinfo(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "USERINFO");
    }

    public void ctcpClientinfo(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "CLIENTINFO");
    }

    public void ctcpErrmsg(String receiver, String text)
        throws IrcMessageError
    {
        this.ctcp(receiver, "ERRMSG " + text);
    }

    public void ctcpPing(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "PING " +
                  String.valueOf(System.currentTimeMillis()));
    }

    public void ctcpTime(String receiver) throws IrcMessageError
    {
        this.ctcp(receiver, "TIME");
    }

}
