package org.thalassocracy.irc;

import java.net.*;
import java.io.*;
import java.util.*;
import java.beans.*;

public class IrcChannel implements IrcEventDispatcher, IrcEventListener {

    private String name;
    BaseIrcClient irc;

    private Vector eventListeners = new Vector();
    public void addEventListener (IrcEventListener listener)
    {
        eventListeners.addElement(listener);
    }
    public void removeEventListener (IrcEventListener listener)
    {
        eventListeners.removeElement(listener);
    }

    private void triggerEvent (IrcEvent foo)
    {
        for (Enumeration e = eventListeners.elements(); e.hasMoreElements();) {
            ((IrcEventListener)e.nextElement()).handleEvent(foo);
        }
    }

    public IrcChannel ()
    {
    }

    public IrcChannel (BaseIrcClient irc, String name)
    {

        this.irc = irc;
        this.name = name;

        irc.addEventListener(this);

    }

    public String getName ()
    {
        return name;
    }

    public void handleEvent(IrcEvent foo)
    {
    }

    public void join(String key) throws IrcMessageError
    {
        irc.join(this.name, key);
    }

    public void join() throws IrcMessageError
    {
        irc.join(this.name);
    }

    public void part() throws IrcMessageError
    {
        irc.part(this.name);
    }

    public void part(String message) throws IrcMessageError
    {
        irc.part(this.name, message);
    }

    public void mode(String modestr) throws IrcMessageError
    {
        irc.chanmode(this.name, modestr);
    }

    public void topic(String topic) throws IrcMessageError
    {
        irc.topic(this.name, topic);
    }

    public void names() throws IrcMessageError
    {
        irc.names(this.name);
    }

    public void invite(String nickname) throws IrcMessageError
    {
        irc.invite(nickname, this.name);
    }

    public void kick(String user, String comment)
        throws IrcMessageError
    {
        irc.kick(this.name, user, comment);
    }

    public void kick(String user) throws IrcMessageError
    {
        irc.kick(this.name, user);
    }

    public void privmsg(String text) throws IrcMessageError
    {
        irc.privmsg(this.name, text);
    }

    public void ctcpAction(String text) throws IrcMessageError
    {
        irc.ctcpAction(this.name, text);
    }

    public Vector getUsers() throws IrcMessageError
    {
        IrcEventWaiter foo = new IrcEventWaiter();
        IrcEvent event;
        IrcRplEvent rplevent = new IrcRplEvent();
        StringTokenExtractor info;
        String nickname;
        Vector users = new Vector();

        synchronized (foo) {
            irc.addEventListener(foo);
            try {
                irc.who(this.name);
                do {
                    foo.wait();
                    event = foo.getEvent();
                    while (event != null) {
                        if (event instanceof IrcRplEvent) {
                            rplevent = (IrcRplEvent)event;
                            if (rplevent.rplnum == 352) {
                                info = new StringTokenExtractor(rplevent.message,
                                                                " ");
                                info.nextToken();
                                info.nextToken();
                                info.nextToken();
                                info.nextToken();
                                info.nextToken();
                                nickname = info.nextToken();
                                if (!nickname.equals("")) {
                                    users.addElement(nickname);
                                }
                            }
                        }
                        event = foo.getEvent();
                    }
                } while (rplevent.rplnum != 315);  /* FIXME - needs timeout */
            } catch (Exception e) { /* FIXME - this should be more specific */
                e.printStackTrace();
            }
        }
        return users;
    }

    public void whoops() throws IrcMessageError
    {
        irc.whoops(this.name);
    }

}
