package org.thalassocracy.irc;

import java.net.*;
import java.io.*;
import java.util.*;
import java.beans.*;

final class EventLoop implements Runnable {

    BaseIrcClient irc;

    public EventLoop(BaseIrcClient irc) throws Exception
    {
        this.irc = irc;
    }

    public void run()
    {
        try {
            for (;;) {
                this.irc.handle_cmd();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

}

public class IrcClient implements IrcEventListener, IrcEventDispatcher {

    private String nick;
    public BaseIrcClient irc;  /* FIXME this needs to be private some day */

    public IrcClient (String host, int port, String nick, String username,
                      String realname) throws Exception
    {
        IrcEventWaiter foo = new IrcEventWaiter();
        IrcEvent event;
        IrcRplEvent rplevent;
        IrcErrRplEvent errrplevent;
        boolean done = false;

        irc = new BaseIrcClient(host, port);

        Thread thread = new Thread(new EventLoop(irc));
        thread.start();

        irc.addEventListener(this);

        synchronized (foo) {
            irc.addEventListener(foo);
            try {
                irc.nick(nick);
                while (!done) {
                    done = true;
                    foo.wait();
                    event = foo.getEvent();
                    if (event instanceof IrcRplEvent) {
                        rplevent = (IrcRplEvent)event;
                        if (rplevent.rplnum == 1) {
                            this.nick = nick;
                        }
                    } else if (event instanceof IrcErrRplEvent) {
                        /* FIXME -- be more fucking selective */
                        System.out.println(((IrcErrRplEvent)event).message);
                        throw new IrcMessageError();
                    } else if (event instanceof IrcPingEvent) {
                        this.handleEvent((IrcPingEvent)event);
                        this.nick = nick;
                    } else {
                        done = false;
                    }
                }
            } catch (Exception e) {  /* maybe this should be more specific */
                e.printStackTrace();
            }
        }

        irc.user(username, host, realname);
    }

    public void addEventListener (IrcEventListener foo)
    {
        irc.addEventListener(foo);
    }

    public void removeEventListener (IrcEventListener foo)
    {
        irc.removeEventListener(foo);
    }

    public String getHost ()
    {
        return irc.getHost();
    }

    public int getPort ()
    {
        return irc.getPort();
    }

    public void setNick (String nick)
    {
        IrcEventWaiter foo = new IrcEventWaiter();
        IrcEvent event;
        IrcNickEvent nickevent;
        boolean done = false;

        synchronized (foo) {
            irc.addEventListener(foo);
            try {
                irc.nick(nick);
                while (!done) {
                    done = true;
                    foo.wait();
                    event = foo.getEvent();
                    if (event instanceof IrcNickEvent) {
                        nickevent = (IrcNickEvent)event;
                        if (nickevent.sender.indexOf(this.nick + "!") == 0) {
                            System.out.println("(setting nick to " + nickevent.nickname + ")");
                            this.nick = nickevent.nickname;
                        }
                    } else if (event instanceof IrcErrRplEvent) {
                        /* FIXME -- be more fucking selective */
                        throw new IrcMessageError();
                    } else {
                        done = false;
                    }
                }
            } catch (Exception e) {  /* maybe this should be more specific */
                e.printStackTrace();
            }
        }
    }

    public String getNick ()
    {
        return this.nick;
    }

    public void handleEvent (IrcEvent foo)
    {
        if (foo instanceof IrcPingEvent) {
            this.handleEvent((IrcPingEvent) foo);
        }
    }

    public void handleEvent (IrcPingEvent foo)
    {
        try {
            irc.pong(foo.server);
        } catch (IrcMessageError e) {
            e.printStackTrace();
        }
    }

    public void handleEvent (IrcNickEvent foo)
    {
    }

}
